from deposit.datasource.abstract_dbsource import AbstractDBSource

from deposit.utils.fnc_files import (is_local_url, url_to_path)
from deposit.utils.fnc_serialize import (GRAPH_ATTRS)

from collections import defaultdict
import json

LEGACY_RESERVED_TABLES = ["#identifier", "#changed", "#local_folder", "#geotags", "#events", "#user_tools", "#queries", "#classless", "#version", "#order"]

class DBRel(AbstractDBSource):
	
	def __init__(self):
		
		AbstractDBSource.__init__(self)
		
		self._current_identifier = None
		self._identifiers = {}  # {identifier: [tablename, ...], ...}
	
	def connect(self):
		
		self._current_identifier = None
		self._identifiers.clear()
		
		return AbstractDBSource.connect(self)
	
	def disconnect(self):
		
		self._current_identifier = None
		self._identifiers.clear()
		
		return AbstractDBSource.disconnect(self)
	
	def is_deposit_db(self):
		
		tables = self.get_tables()
		for name in [
			"identifier", "db_meta", "classes", "subclasses", "relations",
			"user_tools", "queries", "locations",
		]:
			if name not in tables:
				return False
		return True
	
	def is_legacy_db(self):
		
		tables = self.get_tables()
		for name in [
			"#identifier", "#local_folder", "#user_tools",
			"#queries", "#classless", "#version", "#order",
		]:
			if name not in tables:
				return False
		return True
	
	def get_current_identifier(self):
		
		if self._current_identifier is None:
			was_connected = self.is_connected()
			cursor = self.get_cursor()
			identifier_name = "#identifier" if self.is_legacy_db() else "identifier"
			if identifier_name in self.get_tables():
				cursor.execute("SELECT * FROM \"%s\";" % (identifier_name,))
				for row in cursor.fetchall():
					self._current_identifier = json.loads(row[0])
					break
			if not was_connected:
				self.disconnect()
		
		return self._current_identifier
	
	def get_current_tables(self):
		
		return self.get_tables()
	
	def get_identifiers(self):
		
		if (not self._identifiers):
			was_connected = self.is_connected()
			if not was_connected:
				self.connect()
			if self.is_deposit_db() or self.is_legacy_db():
				identifier = self.get_current_identifier()
				if identifier is not None:
					self._identifiers[identifier] = self.get_tables()
			if not was_connected:
				self.disconnect()
		return self._identifiers
	
	def is_valid(self):
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		ret = (cursor is not None) and (self.is_deposit_db() or self.is_legacy_db()) and \
			(self._identifier == self.get_current_identifier())
		
		if not was_connected:
			self.disconnect()
		return ret

	def delete(self):
		
		cursor = self.get_cursor()
		if cursor is None:
			return False
		for table in self.get_tables():
			cursor.execute("DROP TABLE \"%s\";" % (table))
		cursor.connection.commit()
		self.disconnect()
		self.set_identifier(None)
		
		return True
	
	def load_data(self, progress = None):
		# load data from postgres db
		
		if self.is_legacy_db():
			return self.load_legacy_data(progress)
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		tables = self.get_tables()
		
		src_ = GRAPH_ATTRS["source"]
		tgt_ = GRAPH_ATTRS["target"]
		lbl_ = GRAPH_ATTRS["key"]
		id_ = GRAPH_ATTRS["name"]
		relations_ = GRAPH_ATTRS["link"]
		
		data = dict(
			object_relation_graph = dict(
				directed = True,
				multigraph = True,
				graph = {},
				nodes = [],			# [{id_: int, data: dict()}, ...]
				# relations_ = []	# [{src_: int, tgt_: int, lbl_: str, weight: float}, ...]
			),
			# node data = {
			#	(optional) descriptors: {descr_name: value or AbstractDType.to_dict(), ...}
			#	(optional) locations: {descr_name: DGeometry.to_dict(), ...}
			# }
			class_relation_graph = dict(
				directed = True,
				multigraph = True,
				graph = {},
				nodes = [],			# [{id_: str, data: dict()}, ...]
				# relations_ = []	# [{src_: str, tgt_: str, lbl_: str}, ...]
			),
			# node data = {
			#	order: int,
			#	(optional) descriptors = [descr_name, ...]
			# }
			class_membership_graph = dict(
				directed = True,
				multigraph = False,
				graph = {},
				nodes = [],			# [{id_: int/str}, ...]
				# relations_ = []	# [{src_: int/str, tgt_: int/str}, ...]
			),
			local_folder = None,
			max_order = None,
			user_tools = [],
			queries = {},
			deposit_version = None,
		)
		data["object_relation_graph"][relations_] = []
		data["class_relation_graph"][relations_] = []
		data["class_membership_graph"][relations_] = []
		
		table = "db_meta"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		if rows:
			data_table = {}
			for row in rows:
				data_table[row[0]] = json.loads(row[1])
			data["deposit_version"] = data_table.get("deposit_version", None)
			data["local_folder"] = data_table.get("local_folder", None)
			data["max_order"] = data_table.get("max_order", None)
		
		classless_ = None
		pg_class_lookup = {}  # {name: pg_name, ...}
		class_node_lookup = {}  # {name: index in data["class_relation_graph"]["nodes"], ...}
		table = "classes"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			pg_name, name, order = row
			if name is None:
				classless_ = pg_name[1:]
				pg_class_lookup[None] = pg_name
			else:
				pg_class_lookup[name] = pg_name
				class_node_lookup[name] = len(data["class_relation_graph"]["nodes"])
				data["class_relation_graph"]["nodes"].append(
					{id_: name, "data": {"order": order}}
				)
				data["class_membership_graph"]["nodes"].append({id_: name})
		
		table = "subclasses"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			name, subclass = row
			data["class_membership_graph"][relations_].append({src_: name, tgt_: subclass})
		
		pg_relation_lookup = {}  # {(source, relation, target): pg_name, ...}
		table = "relations"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			pg_name, source, relation, target = row
			pg_relation_lookup[(source, relation, target)] = pg_name
			if (source == classless_) or (target == classless_):
				continue
			data["class_relation_graph"][relations_].append(
				{src_: source, tgt_: target, lbl_: relation}
			)
			data["class_relation_graph"][relations_].append(
				{src_: target, tgt_: source, lbl_: "~" + relation}
			)
		
		for key in pg_relation_lookup:
			source, relation, target = key
			cursor.execute("SELECT * FROM \"%s\";" % (pg_relation_lookup[key],))
			rows = cursor.fetchall()
			for row in rows:
				source, target, weight = row
				data_row = {src_: source, tgt_: target, lbl_: relation}
				data_row_rev = {src_: target, tgt_: source, lbl_: "~" + relation}
				if weight is not None:
					data_row[weight] = weight
					data_row_rev[weight] = weight
				data["object_relation_graph"][relations_].append(data_row)
				data["object_relation_graph"][relations_].append(data_row_rev)
		
		obj_node_lookup = {}  # {obj_id: index in data["object_relation_graph"]["nodes"], ...}
		for class_name in pg_class_lookup:
			cursor.execute("SELECT * FROM \"%s\";" % (pg_class_lookup[class_name],))
			descr_names = [desc[0] for desc in cursor.description][1:]
			if class_name is not None:
				data["class_relation_graph"]["nodes"][class_node_lookup[class_name]]["data"]["descriptors"] = descr_names
			rows = cursor.fetchall()
			for row in rows:  # [obj_id, descr, ...]
				obj_id = row[0]
				if class_name is not None:
					data["class_membership_graph"]["nodes"].append({id_: obj_id})
					data["class_membership_graph"][relations_].append(
						{src_: class_name, tgt_: obj_id}
					)
				data_row = {id_: obj_id, "data": {}}
				if descr_names:
					data_row["data"]["descriptors"] = {}
					for i, descr_name in enumerate(descr_names):
						if row[i+1] is not None:
							data_row["data"]["descriptors"][descr_name] = json.loads(row[i+1])
				obj_node_lookup[obj_id] = len(data["object_relation_graph"]["nodes"])
				data["object_relation_graph"]["nodes"].append(data_row)
		
		table = "locations"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			obj_id, descr_name, location = row
			obj_data = data["object_relation_graph"]["nodes"][obj_node_lookup[obj_id]]["data"]
			if "locations" not in obj_data:
				obj_data["locations"] = {}
			obj_data["locations"][descr_name] = json.loads(location)
		
		table = "user_tools"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["user_tools"].append(json.loads(row[0]))
		
		table = "queries"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["queries"][row[0]] = json.loads(row[1])
		
		if not was_connected:
			self.disconnect()
		
		return data
	
	def load_legacy_data(self, progress = None):
		# load data from postgres db (old format)
		
		def load_value(stored_value):
			
			if stored_value.startswith("\u0002"):
				return dict(
					dtype = "DString",
					value = stored_value.lstrip("\u0002"),
				)
			
			if stored_value.startswith('''{"dtype": '''):
				try:
					return json.loads(stored_value)
				except:
					pass
			
			return dict(
				dtype = "DString",
				value = stored_value,
			)
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		tables = self.get_tables()
		
		data = dict(
			classes = {},
			objects = {},
			user_tools = [],
			queries = {},
			local_folder = None,
		)
		'''
		classes[class_name] = {
			order: int,
			objects: [int, ...],
			subclasses: [str, ...],
			descriptors: [str, ...],
			relations: {label_name: [class_name, ...], ...},
		}
		objects[obj_id] = {
			descriptors: {descr_name: {label: descriptor_dict}, ...},
			relations: {label_name: {objects: [int, ...], weights: {obj_id_tgt: weight, ...}, ...}, ...}
		}
		user_tools = [dict(), ...]
		queries[title] = querystr
		'''
		
		cursor.execute("SELECT * FROM \"%s\";" % ("#local_folder",))
		for row in cursor.fetchall():
			local_folder = json.loads(row[0])
			if is_local_url(local_folder):
				local_folder = url_to_path(local_folder)
			data["local_folder"] = local_folder
			break
		
		# load classes
		for table in tables:
			if not table.startswith("#"):
				data["classes"][table] = dict(
					order = None,
					objects = set(),
					subclasses = [],
					descriptors = [],
					relations = defaultdict(list),
				)
		
		# load objects and their classes & descriptors
		descriptor_names = set()
		for table in tables:
			if (not table.startswith("#")) or (table == "#classless"):
				cursor.execute("SELECT * FROM \"%s\";" % (table,))
				descr_names = [desc[0] for desc in cursor.description][1:]
				if table != "#classless":
					for name in descr_names:
						descriptor_names.add(name)
						if name not in data["classes"][table]["descriptors"]:
							data["classes"][table]["descriptors"].append(name)
				rows = cursor.fetchall()
				for row in rows:
					obj_id = row[0]
					if table != "#classless":
						data["classes"][table]["objects"].add(obj_id)
					data["objects"][obj_id] = dict(
						descriptors = {},
						relations = {},
					)
					for i in range(len(descr_names)):
						data["objects"][obj_id]["descriptors"][descr_names[i]] = {"label": load_value(row[i+1])}
		for table in data["classes"]:
			data["classes"][table]["objects"] = list(data["classes"][table]["objects"])
		for name in descriptor_names:
			if name in data["classes"]:
				continue
			data["classes"][name] = dict(
				order = None,
				objects = [],
				subclasses = [],
				descriptors = [],
				relations = [],
			)
		
		# load relations
		for table in tables:
			if table.startswith("#") and (table not in LEGACY_RESERVED_TABLES):
				if table.startswith("##classless#"):
					if table.endswith("##classless"): #  ##classless#rel##classless
						rel = table[12:].split("#")[0]
					else: #  ##classless#rel#cls
						rel, cls2 = table[12:].split("#")
				elif table.endswith("##classless"): #  #cls1#rel##classless
					_, cls1, rel = table[:-11].split("#")
					pass # don't keep class relations to classless
				else: #  #cls1#rel#cls2
					_, cls1, rel, cls2 = table.split("#")
					if cls2 not in data["classes"][cls1]["relations"][rel]:
						data["classes"][cls1]["relations"][rel].append(cls2)
				cursor.execute("SELECT * FROM \"%s\";" % (table,))
				rows = cursor.fetchall()
				for row in rows:
					source_id, target_id = row
					if rel not in data["objects"][source_id]["relations"]:
						data["objects"][source_id]["relations"][rel] = dict(\
							objects = [],
							weights = {},
						)
					if target_id not in data["objects"][source_id]["relations"][rel]["objects"]:
						data["objects"][source_id]["relations"][rel]["objects"].append(target_id)
		for table in data["classes"]:
			data["classes"][table]["relations"] = dict(data["classes"][table]["relations"])
		
		# load class order
		table = "#order"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["classes"][row[0]]["order"] = row[1]
		
		# load user tools
		table = "#user_tools"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["user_tools"].append(json.loads(row[0]))
		
		# load queries
		table = "#queries"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["queries"][row[0]] = json.loads(row[1])
		
		if not was_connected:
			self.disconnect()
		
		return data
	
	def save_data(self, 
		db_meta,
		object_nodes, object_relations, 
		class_nodes, class_relations, 
		member_nodes, member_relations, 
		user_tools, queries, 
		progress = None
	):
		# store data in postgres db
		'''
		db_meta = dict(
			deposit_version,
			local_folder,
			max_order,
		)
		object_nodes = [{id: int, data: object_node_data}, ...]
			object_node_data = {
				(optional) descriptors: {descr_name: value or AbstractDType.to_dict(), ...}
				(optional) locations: {descr_name: DGeometry.to_dict(), ...}
			}
		object_relations = [{src: int, tgt: int, lbl: str, weight: float}, ...]
		class_nodes = [{id: str, data: class_node_data}, ...]
			class_node_data = {
				order: int,
				(optional) descriptors = [descr_name, ...]
			}
		class_relations = [{src: str, tgt: str, lbl: str}, ...]
		member_nodes = [{id: int/str}, ...]
		member_relations = [{src: int/str, tgt: int/str}, ...]
		user_tools = [dict(), ...]
		queries = {name: querystr, ...}
		'''
		
		def create_table(name, columns, cursor):
			
			cursor.execute("CREATE TABLE %s.\"%s\" (%s);" % (self.get_schema(), name, columns))
		
		def save_class(class_name, object_nodes, obj_lookup, class_descriptors, class_order, cursor):
			# object_nodes = {obj_id: data, ...}
			# obj_lookup = {class_name: set(obj_id, ...), ...}
			# class_descriptors = {class_name: set(descriptor_name, ...), ...}
			# class_order = {class_name: order, ...}
			#
			# returns locations = set([(obj_id, descriptor_name, json_data), ...])
			
			locations = set()
			descriptors = class_descriptors[class_name].copy()
			for obj_id in obj_lookup[class_name]:
				descriptors.update(object_nodes[obj_id].get("descriptors", set()))
			obj_column = "id"
			while obj_column in descriptors:
				obj_column += "_"
			descriptors = sorted(descriptors, key = lambda name: class_order[name])
			if descriptors:
				columns = "%s INTEGER, %s" % (obj_column, ", ".join([("\"%s\" TEXT" % name) for name in descriptors]))
			else:
				columns = "%s INTEGER" % (obj_column)
			column_names = ", ".join([("\"%s\"" % name) for name in [obj_column] + descriptors])
			table = "@%s" % (class_name)
			create_table(table, columns, cursor)
			data = [] # [{obj_column: obj_id, [descr_name]: json(value), ...}, ...]
			for obj_id in sorted(obj_lookup[class_name]):
				row = {obj_column: obj_id}
				for name in descriptors:
					row[name] = None
					if ("descriptors" in object_nodes[obj_id]) and (name in object_nodes[obj_id]["descriptors"]):
						row[name] = json.dumps(object_nodes[obj_id]["descriptors"][name])
				if "locations" in object_nodes[obj_id]:
					for name in object_nodes[obj_id]["locations"]:
						locations.add((obj_id, name, json.dumps(object_nodes[obj_id]["locations"][name])))
				data.append(row)
			if data:
				cursor.execute("""
					DROP TYPE IF EXISTS item_;
					CREATE TYPE item_ as (%s);
					INSERT INTO \"%s\" SELECT %s FROM json_populate_recordset(null::item_, %%s);
				""" % (columns, table, column_names), (json.dumps(data),))
			
			return locations	
		
		identifier = self.get_identifier()
		tables = self.get_current_tables()
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		src_ = GRAPH_ATTRS["source"]
		tgt_ = GRAPH_ATTRS["target"]
		lbl_ = GRAPH_ATTRS["key"]
		id_ = GRAPH_ATTRS["name"]
		
		object_nodes = dict([(row[id_], row["data"]) for row in object_nodes])
		# object_nodes = {obj_id: data, ...}
		
		for name in tables:
			cursor.execute("DROP TABLE \"%s\";" % (name))
		
		table = "identifier"
		create_table(table, "name TEXT", cursor)
		cursor.execute("INSERT INTO \"%s\" VALUES ('%s');" % (table, json.dumps(identifier)))
		
		table = "db_meta"
		columns = "variable TEXT, value TEXT"
		create_table(table, columns, cursor)
		if db_meta:
			data = []
			for variable in db_meta:
				data.append({"variable": variable, "value": json.dumps(db_meta[variable])})
			cursor.execute("""
				DROP TYPE IF EXISTS item_;
				CREATE TYPE item_ as (%s);
				INSERT INTO \"%s\" SELECT variable, value FROM json_populate_recordset(null::item_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		class_names = set()
		class_descriptors = {}  # {class_name: set(descriptor_name, ...), ...}
		class_order = {}
		table = "classes"
		columns = "pg_name TEXT, name TEXT, order_ INTEGER"
		create_table(table, columns, cursor)
		data = []
		for row in class_nodes:
			class_names.add(row[id_])
			if "descriptors" in row["data"]:
				class_descriptors[row[id_]] = set(row["data"]["descriptors"])
			else:
				class_descriptors[row[id_]] = set()
			class_order[row[id_]] = row["data"]["order"]
			data.append({"pg_name": "@%s" % (row[id_]), "name": row[id_], "order_": row["data"]["order"]})
		cursor.execute("""
			DROP TYPE IF EXISTS item_;
			CREATE TYPE item_ as (%s);
			INSERT INTO \"%s\" SELECT pg_name, name, order_ FROM json_populate_recordset(null::item_, %%s);
		""" % (columns, table), (json.dumps(data),))
		
		classless_ = "classless_"
		while classless_ in class_names:
			classless_ += "_"
		class_names.add(classless_)
		class_descriptors[classless_] = set()
		
		data = [{"pg_name": "@%s" % (classless_), "name": None, "order_": None}]
		cursor.execute("""
			DROP TYPE IF EXISTS item_;
			CREATE TYPE item_ as (%s);
			INSERT INTO \"%s\" SELECT pg_name, name, order_ FROM json_populate_recordset(null::item_, %%s);
		""" % (columns, table), (json.dumps(data),))
		
		subclasses = []  # [dict(name, subclass), ...]
		class_lookup = defaultdict(set)  # {obj_id: set(class_name, ...), ...}
		obj_lookup = defaultdict(set)  # {class_name: set(obj_id, ...), ...}
		for row in member_relations:
			if isinstance(row[src_], str) and isinstance(row[tgt_], int):
				class_lookup[row[tgt_]].add(row[src_])
				obj_lookup[row[src_]].add(row[tgt_])
			elif isinstance(row[src_], str) and isinstance(row[tgt_], str):
				subclasses.append({"name": row[src_], "subclass": row[tgt_]})
		for obj_id in object_nodes:
			if not class_lookup[obj_id]:
				class_lookup[obj_id].add(classless_)
				obj_lookup[classless_].add(obj_id)
		
		table = "subclasses"
		columns = "name TEXT, subclass TEXT"
		create_table(table, columns, cursor)
		if subclasses:
			cursor.execute("""
				DROP TYPE IF EXISTS item_;
				CREATE TYPE item_ as (%s);
				INSERT INTO \"%s\" SELECT name, subclass FROM json_populate_recordset(null::item_, %%s);
			""" % (columns, table), (json.dumps(subclasses),))
		
		relations = {}  # {pg_name: [dict(source, target, weight), ...], ...}
		relation_classes = []  # [dict(pg_name, source, relation, target), ...]
		for row in class_relations:
			source = row[src_]
			target = row[tgt_]
			label = row[lbl_]
			if label.startswith("~"):
				continue
			pg_name = "@%s_%s_%s" % (source, label, target)
			relations[pg_name] = []
			relation_classes.append({"pg_name": pg_name, "source": source, "relation": label, "target": target})
		for row in object_relations:
			source = row[src_]
			target = row[tgt_]
			label = row[lbl_]
			if label.startswith("~"):
				continue
			weight = row["weight"] if "weight" in row else None
			classes_src = class_lookup[source]
			classes_tgt = class_lookup[target]
			for source_cls in classes_src:
				for target_cls in classes_tgt:
					pg_name = "@%s_%s_%s" % (source_cls, label, target_cls)
					if pg_name not in relations:
						relations[pg_name] = []
						relation_classes.append({"pg_name": pg_name, "source": source_cls, "relation": label, "target": target_cls})
					relations[pg_name].append({"source": source, "target": target, "weight": weight})
		columns = "source INTEGER, target INTEGER, weight FLOAT"
		for table in relations:
			create_table(table, columns, cursor)
			cursor.execute("""
				DROP TYPE IF EXISTS item_;
				CREATE TYPE item_ as (%s);
				INSERT INTO \"%s\" SELECT source, target, weight FROM json_populate_recordset(null::item_, %%s);
			""" % (columns, table), (json.dumps(relations[table]),))
		
		table = "relations"
		columns = "pg_name TEXT, source TEXT, relation TEXT, target TEXT"
		create_table(table, columns, cursor)
		cursor.execute("""
			DROP TYPE IF EXISTS item_;
			CREATE TYPE item_ as (%s);
			INSERT INTO \"%s\" SELECT pg_name, source, relation, target FROM json_populate_recordset(null::item_, %%s);
		""" % (columns, table), (json.dumps(relation_classes),))
		
		locations = set()  # [(obj_id, descriptor_name, json_data), ...]
		for class_name in class_names:
			locations.update(save_class(class_name, object_nodes, obj_lookup, class_descriptors, class_order, cursor))
		
		table = "locations"
		columns = "id INTEGER, descriptor TEXT, location TEXT"
		create_table(table, columns, cursor)
		data = []
		for obj_id, descriptor_name, location in locations:
			data.append({"id": obj_id, "descriptor": descriptor_name, "location": location})
		if data:
			cursor.execute("""
				DROP TYPE IF EXISTS item_;
				CREATE TYPE item_ as (%s);
				INSERT INTO \"%s\" SELECT id, descriptor, location FROM json_populate_recordset(null::item_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = "user_tools"
		columns = "data TEXT"
		create_table(table, columns, cursor)
		if user_tools:
			data = [dict(data = json.dumps(row)) for row in user_tools]
			cursor.execute("""
				DROP TYPE IF EXISTS item_;
				CREATE TYPE item_ as (%s);
				INSERT INTO \"%s\" SELECT data FROM json_populate_recordset(null::item_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = "queries"
		columns = "title TEXT, querystr TEXT"
		create_table(table, columns, cursor)
		if queries:
			data = []
			for title in queries:
				data.append(dict(title = title, querystr = json.dumps(queries[title])))
			cursor.execute("""
				DROP TYPE IF EXISTS item_;
				CREATE TYPE item_ as (%s);
				INSERT INTO \"%s\" SELECT title, querystr FROM json_populate_recordset(null::item_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		cursor.connection.commit()
		
		if not was_connected:
			self.disconnect()
		
		return True




from deposit.store.abstract_dtype import AbstractDType
import datetime

class DDateTime(AbstractDType):
	
	@property
	def isoformat(self):
		
		value = AbstractDType.get_value(self)
		if isinstance(value, str):
			return value
		if isinstance(value, datetime.datetime):
			return value.isoformat()
		return None
	
	def get_value(self):
		
		value = AbstractDType.get_value(self)
		if isinstance(value, str):
			AbstractDType.set_value(self, datetime.datetime.fromisoformat(value))
		return AbstractDType.get_value(self)
	
	def set_value(self, value):
		
		if isinstance(value, DDateTime):
			AbstractDType.set_value(self, value.value)
			return
		AbstractDType.set_value(self, value)
	
	def to_dict(self):
		
		value = AbstractDType.get_value(self)
		if isinstance(value, datetime.datetime):
			value = value.isoformat()
		
		return dict(
			dtype = self.__class__.__name__,
			value = value,
		)
	
	def __str__(self):
		
		value = AbstractDType.get_value(self)
		if isinstance(value, datetime.datetime):
			value = value.isoformat()
		
		return "%s(%s)" % (self.__class__.__name__, value)



'''
Query
	.querystr
	.columns = [(class_name, descriptor_name), (None, alias), ...]
	.classes = [class_name, ...]
	.objects = {obj_id, ...}
	.main_class = class_name
	.hash
	.parse
		.classes = {name, ...}
		.descriptors = {name, ...}
		.querystr
		.columns = [(class_name, descriptor_name), (None, alias), ...]
		.selects = [(class_name, descriptor_name), ...]
		.group_by = [(class_name, descriptor_name), ...]
		.counts = [(alias, class_name, descriptor_name), ...]
		.sums = [(alias, class_name, descriptor_name), ...]
		.relations = [(class1, relation, class2), ...]
		.where_expr = python expression
		.where_vars = {name: (class, descriptor), ...}
			use: eval(where_expr, {name: value of descriptor of object which is member of class, ...})
		.classes_used = [name, ...]

Query[row, column] / [row, Class_name, Descriptor_name] / [row, "Class_name.Descriptor_name"] = (obj_id, value)

value = python type / DDateTime / DGeometry / DResource

DDateTime
	.value = datetime.datetime
	.isoformat = value in ISO format

DGeometry
	.value = (geometry_type, coords, srid, srid_vertical)
	.geometry_type = POINT etc.
	.coords = []
	.srid
	.srid_vertical
	.wkt = WKT string

DResource
	.value = (url, filename, is_stored, is_image)
	.url
	.filename
	.is_stored
	.is_image
	.object_ids = set()
	
	
'''

from deposit.query.parse import Parse
from deposit.store.abstract_dtype import AbstractDType

from collections import defaultdict
from natsort import (natsorted)
import traceback
import sys

class Query(object):
	
	def __init__(self, store, querystr, progress = None):
		
		self.querystr = querystr
		self.parse = None
		self.hash = [] # hash of the query rows (to quickly check whether query has changed)
		
		self._store = store
		self._rows = []
		self._columns = []
		self._classes = []
		self._objects = set()
		self._main_class = None
		self._progress = progress
		
		self.process()
	
	def set_progress(self, progress):
		
		self._progress = progress
	
	def _get_descriptor_value(self, obj_id, class_name, descriptor_name, class_lookup, descr_lookup, obj_if_none = False):
		
		if class_name not in class_lookup[obj_id][0]:
			return None
		
		if descriptor_name is None:
			if obj_if_none:
				return obj_id
			return None
		
		return descr_lookup[obj_id].get((class_name, descriptor_name), None)
	
	def _get_order(self, class_name, descriptor_name):
		
		order = []
		for name in [class_name, descriptor_name]:
			if self._store.has_class(name):
				order.append(self._store.get_class(name).order)
			else:
				order.append(None)
		return order
	
	def process(self):
		
		self._process()  # DEBUG
		return  # DEBUG
		try:
			self._process()
		except:
			print("QUERY ERROR:", sys.exc_info())
	
	def _process(self):
		
		self.hash = []
		self._rows = []
		self._columns = []
		self._column_idxs = {}
		self._classes = []
		self._objects = set()
		self._main_class = None
		
		self.parse = Parse(
			self.querystr,
			set([cls.name for cls in self._store.get_classes()]),
			set(self._store.get_descriptor_names())
		)
		
		selects = []
		for class_name, descriptor_name in self.parse.selects:
			if (class_name not in self.parse.classes) and (class_name not in ["*", "!*"]) and (not isinstance(class_name, tuple)):
				continue
			if (descriptor_name not in self.parse.descriptors) and (descriptor_name not in [None, "*"]):
				continue
			selects.append((class_name, descriptor_name))
		
		if not selects:
			return
		
		self._main_class = selects[0][0]
		
		keys = set()  # {(class_name, descriptor_name), ...}
		for class_name, descriptor_name in selects:
			keys.add((class_name, descriptor_name))
			if class_name not in self._classes:
				self._classes.append(class_name)
			if descriptor_name not in self._classes:
				self._classes.append(descriptor_name)
		alias_lookup = {}
		for alias, class_name, descriptor_name in self.parse.counts:
			keys.add((class_name, descriptor_name))
			alias_lookup[alias] = ("count", class_name, descriptor_name)
			if class_name not in self._classes:
				self._classes.append(class_name)
		for alias, class_name, descriptor_name in self.parse.sums:
			keys.add((class_name, descriptor_name))
			alias_lookup[alias] = ("sum", class_name, descriptor_name)
			if class_name not in self._classes:
				self._classes.append(class_name)
		value_lookup = {}
		for name in self.parse.where_vars:
			value_lookup[self.parse.where_vars[name]] = name
		key_lookup = defaultdict(set)
		
		# get connected objects based on selects and relations
		paths = self._store.get_class_connections(
			self.parse.classes_used, 
			self.parse.relations, 
			progress = self._progress
		)
		
		obj_ids = set()
		for path in paths:
			obj_ids.update(path)
		
		cmax = len(paths) + len(obj_ids)
		cnt = 1
		
		class_lookup = {}  # {obj_id: [classes, superclasses], ...}
		class_descr_lookup = {"!*": [], "*": []}
		for obj_id in obj_ids:
			class_lookup[obj_id] = [set(), set()]
		for class_name in self._classes:
			class_descr_lookup[class_name] = []
			cls = self._store.get_class(class_name)
			if cls is not None:
				class_descr_lookup[class_name] = cls.get_descriptor_names()
				for tgt in self._store.G.iter_class_children(cls.name):
					if tgt in class_lookup:
						class_lookup[tgt][0].add(cls.name)
				for tgt in self._store.G.iter_class_descendants(cls.name):
					if tgt in class_lookup:
						class_lookup[tgt][1].add(cls.name)
		for obj_id in obj_ids:
			if not class_lookup[obj_id][0]:
				class_lookup[obj_id][0].add(None)
		
		descr_lookup = {}  # {obj_id: {(class_name, descriptor_name): value, ...}, ...}
		class_names = set()
		for obj_id in obj_ids:
			if (self._progress is not None) and (cnt % 20000 == 0):
				self._progress.update_state(text = "Populating Query", value = cnt, maximum = cmax)
			cnt += 1
			descr_lookup[obj_id] = {}
			obj = self._store.G.get_object_data(obj_id)
			for descr in obj._descriptors:
				classes = class_lookup[obj_id][0].union(class_lookup[obj_id][1])
				if classes:
					for class_name in classes:
						descr_lookup[obj_id][(class_name, descr.name)] = obj._descriptors[descr]
				else:
					descr_lookup[obj_id][(None, descr.name)] = obj._descriptors[descr]
		
		rows = []  # [(obj, {key: value, ...}), ...]; key = (class_name, descriptor_name)
		done = set()  # {(obj_id, ...), ...}; to avoid duplicate rows
		keys_collect = set()  # {(class_name, descriptor_name), ...}
		for path in paths:
			if (self._progress is not None) and (cnt % 20000 == 0):
				self._progress.update_state(text = "Populating Query", value = cnt, maximum = cmax)
			cnt += 1
			# filter paths by evaluating WHERE expression for each
			values = {}  # {name: value, ...}
			if self.parse.where_expr:
				# collect descriptor values for each path object according to WHERE (Parse.where_vars)
				for name in self.parse.where_vars:
					values[name] = None
				for obj_id in path:
					class_names = class_lookup[obj_id][0]
					for name in self.parse.where_vars:
						class_name, descriptor_name = self.parse.where_vars[name]
						if class_name == "!*":
							class_name = None
						if (not isinstance(class_name, tuple)) and class_names and (class_name not in class_names):
							continue
						value = self._get_descriptor_value(obj_id, class_name, descriptor_name, class_lookup, descr_lookup, obj_if_none = True)
						if value is not None:
							values[name] = value
				if not eval(self.parse.where_expr, values):
					continue
			
			# collect descriptor values according to SELECT, COUNT, SUM
			data = {}  # {key: value, ...}; key = (class_name, descriptor_name)
			for obj_id in path:
				self._objects.add(obj_id)
				class_names = class_lookup[obj_id][1]
				for key in keys:
					class_name, descriptor_name = key
					if (not isinstance(class_name, tuple)) and (class_name[-1] != "*") and (class_name not in class_names):
						continue
					if isinstance(class_name, tuple) and (obj_id not in list(class_name)):
						continue
					if class_name == "*":
						class_name = class_names[0] if class_names else None
					elif class_name == "!*":
						class_name = None
					if descriptor_name != "*":
						data[(class_name, descriptor_name)] = (obj_id, self._get_descriptor_value(obj_id, class_name, descriptor_name, class_lookup, descr_lookup))
						key_lookup[key].add((class_name, descriptor_name))
					else:
						descriptor_names = class_descr_lookup["!*" if class_name is None else class_name]
						for _, descriptor_name in descr_lookup[obj_id]:
							if descriptor_name not in descriptor_names:
								descriptor_names.append(descriptor_name)
						if descriptor_names:
							for descriptor_name in descriptor_names:
								data[(class_name, descriptor_name)] = (obj_id, self._get_descriptor_value(obj_id, class_name, descriptor_name, class_lookup, descr_lookup))
								key_lookup[key].add((class_name, descriptor_name))
						else:
							data[(class_name, None)] = (obj_id, None)
							key_lookup[key].add((class_name, None))
			
			check_row = tuple(sorted([val[0] for val in data.values()]))
			if check_row in done:
				continue
			done.add(check_row)
			rows.append((path[0], data))
			keys_collect.update(data.keys())
		
		# prune (class_name, None) from key_lookup if it already has (class_name, descriptor_name)
		for key in key_lookup:
			for class_name, descriptor_name in list(key_lookup[key]):
				if descriptor_name is None:
					for class_name2, descriptor_name2 in key_lookup[key]:
						if descriptor_name2 is None:
							continue
						if class_name2 == class_name:
							key_lookup[key].remove((class_name, None))
							break
		
		# generate keys for each column sorted according to Parse.columns
		keys_columns = []  # [(typ, alias, class_name, descriptor_name), ...]; typ = count / sum / None
		for class_name, descriptor_name in self.parse.columns:
			if class_name is None:
				if descriptor_name in alias_lookup:
					alias = descriptor_name
					typ_, class_name_, descriptor_name_ = alias_lookup[alias]
					if (class_name_, descriptor_name_) in keys_collect:
						keys_columns.append((typ_, alias, class_name_, descriptor_name_))
			elif (class_name, descriptor_name) in key_lookup:
				for class_name_, descriptor_name_ in natsorted(list(key_lookup[(class_name, descriptor_name)]), key = lambda row: self._get_order(*row)):
					if (class_name_, descriptor_name_) in keys_collect:
						keys_columns.append((None, None, class_name_, descriptor_name_))
		idxs_counts = []
		idxs_sums = []
		for idx, key in enumerate(keys_columns):
			typ, _, _, _ = key
			if typ == "count":
				idxs_counts.append(idx)
			elif typ == "sum":
				idxs_sums.append(idx)
		
		# group rows according to GROUP BY
		keys_group = [key for key in self.parse.group_by if key in keys_collect]
		grouped = defaultdict(list)
		for obj, data in rows:
			row = []
			group_key = []
			for key in keys_columns:
				typ, alias, class_name, descriptor_name = key
				row.append(data.get((class_name, descriptor_name), (None, None)))
				if (class_name, descriptor_name) in keys_group:
					group_key.append(str(row[-1]))
			row = tuple([obj] + row)
			if group_key:
				group_key = tuple(group_key)
			else:
				group_key = tuple([str(val) for val in row])
			grouped[group_key].append(row)
		
		# add query rows and count / sum grouped
		rows = []
		for group_key in grouped:
			if len(grouped[group_key]) == 1:
				row_collect = list(grouped[group_key].pop())
				obj = row_collect.pop(0)
				for idx in idxs_counts:
					row_collect[idx] = (None, int(row_collect[idx] is not None))
				for idx in idxs_sums:
					val = 0
					try:
						val = float(row_collect[idx])
					except:
						pass
					row_collect[idx] = (None, val)
			else:
				row_collect = [None] * len(keys_columns)
				for row in grouped[group_key]:
					row = list(row)
					obj = row.pop(0)
					for idx, key in enumerate(keys_columns):
						if idx in idxs_counts:
							if row_collect[idx] is None:
								row_collect[idx] = [None, 0]
							if row[idx] is not None:
								row_collect[idx][1] += 1
						elif idx in idxs_sums:
							if row_collect[idx] is None:
								row_collect[idx] = [None, 0]
							val = None
							try:
								val = float(row[idx][1])
							except:
								pass
							if val is not None:
								row_collect[idx][1] += val
						elif row_collect[idx] is None:
							row_collect[idx] = row[idx]
			rows.append([tuple(item) for item in row_collect])
		
		# sort and prune rows
		rows = natsorted(rows)[::-1]
		self._rows = []
		check_row_last = None
		for row in rows:
			check_row = set([item[0] if (item[0] is not None) else item[1] for item in row if item != (None, None)])
			if check_row_last is None:
				check_row_last = check_row
			elif check_row.issubset(check_row_last):
				continue
			check_row_last = check_row
			self._rows.append(row)
		self._rows = self._rows[::-1]
		
		for _, alias, class_name, descriptor_name in keys_columns:
			if alias is None:
				self._columns.append((class_name, descriptor_name))
			else:
				self._columns.append((None, alias))
		
		for idx, key in enumerate(self._columns):
			self._column_idxs[key] = idx
		
#		self._rows = [[value, ...]]
#		self._columns = [(class_name, descriptor_name) or (None, alias)]
#		self._classes = [class_name, ...]
	
	@property
	def columns(self):
		# [(class_name, descriptor_name), (None, alias), ...]
		
		return self._columns.copy()
	
	@property
	def classes(self):
		# [class_name, ...]
		
		return self._classes.copy()
	
	@property
	def objects(self):
		# [obj_id, ...]
		
		return self._objects
	
	@property
	def main_class(self):
		
		return self._main_class
	
	def __len__(self):

		return len(self._rows)
	
	def __bool__(self):

		return len(self._rows) > 0
	
	def __getitem__(self, idx):
		# idx = (row, column) / (row, Class_name, Descriptor_name) / (row, "Class.Descriptor")
		#
		# returns (obj_id, value)
		
		if not isinstance(idx, tuple):
			raise IndexError()
		if not isinstance(idx[0], int):
			raise IndexError()
		row = idx[0]
		col = None
		if (len(idx) == 2) and isinstance(idx[1], int):
			col = idx[1]
		elif (len(idx) == 2) and isinstance(idx[1], str):
			col = tuple(idx[1].split("."))
			if col in self._column_idxs:
				col = self._column_idxs[col]
		elif len(idx) == 3:
			if idx[1:] in self._column_idxs:
				col = self._column_idxs[idx[1:]]
		if col is None:
			return (None, None)
		
		return self._rows[row][col]
	
	def __iter__(self):
		
		for row in self._rows:
			yield row
	
	def __setitem__(self, idx, row):

		raise Exception("Impossible to alter a Query")
	
	def __delitem__(self, idx):

		raise Exception("Impossible to alter a Query")




from deposit import __version__
from deposit.datasource.abstract_filesource import AbstractFileSource
from deposit.store.abstract_delement import AbstractDElement
from deposit.store.dobject import DObject
from deposit.store.dclass import DClass

from deposit.utils.fnc_serialize import (json_data_to_store, GRAPH_ATTRS)

import json

class DJSONEncoder(json.JSONEncoder):
	
	def default(self, o):
		
		if isinstance(o, AbstractDElement):
			element_data = o.to_dict()
			return element_data
		return o.__dict__

class JSON(AbstractFileSource):
	
	EXTENSION = "json"
	
	def save_data(self, store, resources, path):
		
		self.update_progress(1, 10, text = "Saving data")
		
		with open(path, "w") as f:
			json.dump(dict(
				object_relation_graph = store.G.objects_to_json(GRAPH_ATTRS),
				class_relation_graph = store.G.classes_to_json(GRAPH_ATTRS),
				class_membership_graph = store.G.members_to_json(GRAPH_ATTRS),
				local_folder = store._local_folder,
				max_order = store._max_order,
				user_tools = store._user_tools,
				queries = store._queries,
				deposit_version = __version__,
			), f, cls = DJSONEncoder)
		
		self.update_progress(10)
		
		return True
	
	def load_data(self, path):
		
		with open(path, "r") as f:
			data = json.load(f)
		return data
	
	def data_to_store(self, data, store):
		
		return json_data_to_store(data, store, self._progress)



version_info = (1, 4, 1)

__version__ = '.'.join(map(str, version_info))
__title__ = 'Deposit'
__date__ = "23.08.2022"

from deposit.store.store import Store

from deposit.store.ddatetime import DDateTime
from deposit.store.dgeometry import DGeometry
from deposit.store.dresource import DResource



from deposit.store.abstract_dtype import AbstractDType

class DResource(AbstractDType):
	
	def __init__(self, value = None):
		
		AbstractDType.__init__(self, value)
		
		self._objects = set()
	
	@property
	def url(self):
		
		return self.get_value()[0]
	
	@property
	def filename(self):
		
		return self.get_value()[1]
		
	@property
	def is_stored(self):
		
		return (self.get_value()[2] == True)
	
	@property
	def is_image(self):
		
		return (self.get_value()[3] == True)
	
	@property
	def object_ids(self):
		# ids of Objects which have resource as a Descriptor
		
		return self._objects
	
	def get_value(self):
		# return (url, filename, is_stored, is_image)
		
		value = AbstractDType.get_value(self)
		if value is None:
			AbstractDType.set_value(self, (None, None, False, False))
		else:
			pass
		return AbstractDType.get_value(self)
	
	def set_value(self, value):
		# value = (url / path, filename, is_stored, is_image)
		
		if value is None:
			AbstractDType.set_value(self, (None, None, False, False))
			return
		
		if (not isinstance(value, tuple)) or (len(value) != 4):
			raise Exception("Invalid value specified: %s" % (str(value)))
		
		AbstractDType.set_value(self, value)
	
	def add_object(self, obj_id):
		
		self._objects.add(obj_id)
	
	def del_object(self, obj_id):
		
		if obj_id in self._objects:
			self._objects.remove(obj_id)
	
	def to_dict(self):
		
		return dict(
			dtype = self.__class__.__name__,
			value = self._value,
			objects = list(self._objects),
		)
	
	def from_dict(self, data):
		
		self._value = data["value"]
		self._objects = set(data["objects"])
		
		return self
	
	def __str__(self):
		
		url, _, _, _ = AbstractDType.get_value(self)
		
		return "%s(%s)" % (self.__class__.__name__, str(url))



from deposit.store.abstract_delement import AbstractDElement
from deposit.store.ddatetime import DDateTime
from deposit.store.dgeometry import DGeometry
from deposit.store.dresource import DResource

from deposit.utils.fnc_geometry import (coords_to_wkt, wkt_to_coords)

class DObject(AbstractDElement):
	
	def __init__(self, store, obj_id):
		
		AbstractDElement.__init__(self, store)
		
		self.id = obj_id
		
		self._descriptors = {}  # {DClass: value, ...}
		self._locations = {}  # {DClass: DGeometry, ...}
		self._json_descriptors = None
		self._json_locations = None
	
	def _on_class_deleted(self, cls):
		
		if cls in self._descriptors:
			del self._descriptors[cls]
			self._store.callback_changed([self])
			return True
		return False
	
	def to_dict(self):
		
		from deposit.utils.fnc_serialize import (dtype_to_dict)
		
		descriptors = dict([(descr.name, dtype_to_dict(self._descriptors[descr])) for descr in self._descriptors])
		locations = dict([(descr.name, self._locations[descr].to_dict()) for descr in self._locations])
		data = {}
		if descriptors:
			data["descriptors"] = descriptors
		if locations:
			data["locations"] = locations
		
		return data
	
	def from_dict_1(self, data):
		
		self._json_descriptors = data.get("descriptors", {})
		self._json_locations = data.get("locations", {})
		
		for name in self._json_descriptors:
			value = self._json_descriptors[name]
			if isinstance(value, dict) and (value["dtype"] == "DResource"):
				url, filename, is_stored, is_image = value["value"]
				self._json_descriptors[name] = self._store.add_resource(url, filename, is_stored, is_image)
				self._json_descriptors[name].add_object(self.id)
		
		return self
	
	def from_dict_2(self):
		
		from deposit.utils.fnc_serialize import (value_to_dtype)
		
		for name in self._json_descriptors:
			self._descriptors[self._store.get_class(name)] = value_to_dtype(self._json_descriptors[name])
		self._json_descriptors = None
		for name in self._json_locations:
			self._locations[self._store.get_class(name)] = DGeometry().from_dict(self._json_locations[name])
		self._json_locations = None
	
	
	# ---- Descriptors -------------------------------------------------------
	def set_descriptor(self, name, value):
		
		if value == self.get_descriptor(name):
			return
		self.del_descriptor(name)
		if value is None:
			return
		descr = self._store.add_class(name)
		self._descriptors[descr] = value
		if isinstance(value, DResource):
			self._descriptors[descr].add_object(self.id)
		for cls in self.get_classes():
			cls.set_descriptor(name)
		self._store._on_descriptor_set(descr)
		self._store.callback_changed([self])
	
	def set_resource_descriptor(
		self, name, url, filename = None, is_stored = None, is_image = None
	):
		# filename = "name.ext"; if None, determine automatically
		# is_stored = True if resource is stored in local folder
		#	= False, not stored locally
		#	= None, attempt to store it locally if possible & set is_stored flag
		# is_image = True if resource is an image; if None, determine automatically
		
		value = self._store.add_resource(url, filename, is_stored, is_image)
		value.add_object(self.id)
		self.set_descriptor(name, value)
	
	def set_geometry_descriptor(
		self, name, geometry_type, coords, srid = None, srid_vertical = None
	):
		# geometry_type = str (POINT / MULTIPOINT / LINESTRING / POLYGON / MULTIPOLYGON / POINTZ etc. / POINTM etc.)
		# coords = POINT: [x,y]
		# 	POINTZ: [x, y, z] etc.
		#	POINTM: [x, y, z, m] etc.
		#	MULTIPOINT / LINESTRING: [[x, y], ...]
		#	POLYGON: [exterior, hole1, hole2, ...]; exterior / hole = [[x, y], ...]
		#	MULTIPOLYGON: [POLYGON, ...]
		
		self.set_descriptor(name, DGeometry((geometry_type, coords, srid, srid_vertical)))
	
	def set_datetime_descriptor(self, name, value):
		# value = datetime.datetime or str in ISO format
		
		self.set_descriptor(name, DDateTime(value))
	
	def get_descriptor(self, name, default = None):
		
		descr = self._store.get_class(name)
		if descr is None:
			return default
		if descr in self._descriptors:
			return self._descriptors[descr]
		return default
	
	def has_descriptors(self):
		
		return len(self._descriptors) > 0
	
	def has_descriptor(self, name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return False
		return (descr in self._descriptors)
	
	def get_descriptors(self, ordered = False):
		
		if ordered:
			return sorted(list(self._descriptors.keys()), key = lambda descr: descr.order)
		return list(self._descriptors.keys())
	
	def get_descriptor_names(self, ordered = False):
		
		return [descr.name for descr in self.get_descriptors(ordered)]
	
	def rename_rescriptor(self, name, new_name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return
		if descr not in self._descriptors:
			return
		value = self._descriptors[descr]
		del self._descriptors[descr]
		descr = self._store.add_class(new_name)
		self._descriptors[descr] = value
		for cls in self.get_classes():
			cls.set_descriptor(new_name)
		self._store._on_descriptor_deleted()
		self._store.callback_changed([self])
	
	def del_descriptor(self, name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return
		if descr in self._descriptors:
			resource = None
			if isinstance(self._descriptors[descr], DResource):
				resource = self._descriptors[descr]
			del self._descriptors[descr]
			if descr in self._locations:
				del self._locations[descr]
			self._store._on_descriptor_deleted()
			if (resource is not None) and (resource not in self._descriptors.values()):
				resource.del_object(self.id)
				self._store._on_resource_deleted(resource)
			self._store.callback_changed([self])
	
	
	# ---- Locations ---------------------------------------------------------
	def set_location(self, name, value):
		# set location of the object on an image resource by a geometric shape ([0,0] being the bottom-left corner)
		# value = DGeometry
		#	WKT or 
		# 	(geometry_type, coords, srid, srid_vertical) or
		#	(geometry_type, coords, srid) or
		#	(geometry_type, coords)
		
		if value is None:
			self.del_location(name)
			return
		descr = self._store.add_class(name)
		if descr not in self._descriptors:
			raise Exception("Could not set location. Descriptor %s of Object %d does not exist" % (name, self.id))
		self._locations[descr] = DGeometry(value)
		self._store.callback_changed([self])
	
	def get_location(self, name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return None
		if descr in self._locations:
			return self._locations[descr]
		return None
	
	def del_location(self, name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return
		if descr in self._locations:
			del self._locations[descr]
			self._store.callback_changed([self])
	
	
	# ---- Classes -----------------------------------------------------------
	def get_classes(self, ordered = False, superclasses = False):
		# returns [DClass, ...]
		
		classes = set()
		if superclasses:
			for src in self._store.G.iter_class_ancestors(self.id):
				if isinstance(src, str):
					classes.add(self._store.G.get_class_data(src))
		else:
			for src in self._store.G.iter_class_parents(self.id):
				classes.add(self._store.G.get_class_data(src))
		
		if ordered:
			return sorted(list(classes), key = lambda cls: cls.order)
		return list(classes)
	
	def has_class(self, name = None):
		
		if name.__class__.__name__ == "DClass":
			name = name.name
		
		for src in self._store.G.iter_class_parents(self.id):
			if name is None:
				return True
			if src == name:
				return True
		return False
	
	
	# ---- Relations ---------------------------------------------------------
	def add_relation(self, obj_id_tgt, label, weight = None):
		
		if isinstance(obj_id_tgt, DObject):
			obj_id_tgt = obj_id_tgt.id
		self._store.check_obj_id(obj_id_tgt)
		obj_id_src = self.id
		obj_id_src, obj_id_tgt, label = self._store._check_inverse_rel(obj_id_src, obj_id_tgt, label)
		self._store.G.add_object_relation(obj_id_src, obj_id_tgt, label, weight)
		self._store.G.add_object_relation(obj_id_tgt, obj_id_src, "~" + label, weight)
		self._store._on_relation_added(label)
		obj = self._store.G.get_object_data(obj_id_tgt)
		self._store.callback_changed([self, obj])
	
	def has_relation(self, obj_id_tgt, label):
		
		if isinstance(obj_id_tgt, DObject):
			obj_id_tgt = obj_id_tgt.id
		return self._store.G.has_object_relation(self.id, obj_id_tgt, label)
	
	def get_relations(self, obj_id_tgt = None):
		# returns [(DObject, label), ...]
		
		if isinstance(obj_id_tgt, DObject):
			obj_id_tgt = obj_id_tgt.id
		
		for tgt, label in self._store.G.iter_object_relations(self.id):
			if (obj_id_tgt is not None) and (tgt != obj_id_tgt):
				continue
			yield (self._store.G.get_object_data(tgt), label)
	
	def get_relation_weight(self, obj_id_tgt, label):
		
		if isinstance(obj_id_tgt, DObject):
			obj_id_tgt = obj_id_tgt.id
		if self._store.G.has_object_relation(self.id, obj_id_tgt, label):
			return self._store.G.get_object_relation_weight(self.id, obj_id_tgt, label)
		return None
	
	def del_relation(self, obj_id_tgt, label):
		
		if isinstance(obj_id_tgt, DObject):
			obj_id_tgt = obj_id_tgt.id
		obj_id_src = self.id
		obj_id_src, obj_id_tgt, label = self._store._check_inverse_rel(obj_id_src, obj_id_tgt, label)
		
		if self._store.G.has_object_relation(obj_id_src, obj_id_tgt, label):
			self._store.G.del_object_relation(obj_id_src, obj_id_tgt, label)
			self._store.G.del_object_relation(obj_id_tgt, obj_id_src, "~" + label)
			self._store._on_relation_deleted()
			obj = self._store.G.get_object_data(obj_id_tgt)
			self._store.callback_changed([self, obj])
	
	
	def __repr__(self):
		
		return "DObject(%d)" % (self.id)




from deposit.externalsource.abstract_externalsource import AbstractExternalsource
from deposit.externalsource.xlsx import XLSX
from deposit.externalsource.csv import CSV
from deposit.externalsource.shp import SHP



from deposit.datasource.abstract_datasource import AbstractDatasource
from deposit.store.dresource import DResource

from deposit.utils.fnc_files import (as_url, url_to_path, copy_resources)
from deposit.utils.fnc_serialize import (legacy_data_to_store)

import os

class AbstractFileSource(AbstractDatasource):
	
	EXTENSION = "file"
	
	def __init__(self):
		
		AbstractDatasource.__init__(self)
		
		self._path = None
		self._progress = None
	
	def to_dict(self):
		
		data = AbstractDatasource.to_dict(self)
		data["url"] = self.get_url()
		
		return data
	
	def get_name(self):
		
		name = self.get_url()
		if name is not None:
			name = os.path.basename(name)
			if name:
				return os.path.splitext(name)[0]
		return "file"
	
	def is_valid(self):
		
		return os.path.isfile(self._path)
	
	def can_create(self):
		# return error / warning code
		
		path = self.get_path()
		if path is None:
			return False
		
		return True
	
	def get_folder(self):
		
		if self._path is not None:
			return os.path.dirname(self._path)
		return None
	
	def get_url(self):
		
		if self._path is not None:
			return as_url(self._path)
		return None
	
	def get_path(self):
		
		return self._path
	
	def set_path(self, path):
		
		self._path = os.path.normpath(os.path.abspath(path))
	
	def create(self):
		
		folder = self.get_folder()
		if folder is None:
			return False
		
		if not os.path.isdir(folder):
			try:
				os.makedirs(folder, exist_ok = True)
			except:
				return False
		
		return True
	
	def save_data(self, store, resources, path):
		# re-implement
		
		return {}
	
	def save(self, store, progress = None, path = None, url = None, *args, **kwargs):
		
		self.set_progress(progress)
		
		if (path is None) and (url is not None):
			path = url_to_path(url)
		
		if path == self.get_path():
			path = None
		
		if path is not None:
			self.set_path(path)
			if not self.create():
				return False
		
		path = self.get_path()
		path, ext = os.path.splitext(path)
		if ext.lower() != ".%s" % (self.EXTENSION):
			ext = ".%s" % (self.EXTENSION)
		path = path + ext
		
		# copy all is_stored resources to new folder (if not already there)
		src_folder = store.get_folder()
		if store.has_local_folder():
			dst_folder = src_folder
		else:
			dst_folder = os.path.normpath(os.path.abspath(os.path.dirname(path)))
		resources = {}
		if dst_folder == src_folder:
			resources = store._resources
		elif store._resources:
			cmax = len(store._resources)
			cnt = 1
			self.update_progress(cnt, cmax, text = "Copying files")
			resources = copy_resources(
				store._resources, src_folder, dst_folder, progress = self._progress
			)
			if not resources:
				return False
		
		return self.save_data(store, resources, path)
	
	def load_data(self, path):
		# re-implement
		
		return {}
	
	def load(self, store, progress = None, path = None, url = None, *args, **kwargs):
		
		self.set_progress(progress)
		
		if (path is None) and (url is not None):
			path = url_to_path(url)
		
		if path == self.get_path():
			path = None
		
		if path is not None:
			self.set_path(path)
			if not self.is_valid():
				return False
		
		path = self.get_path()
		if path is None:
			print("LOAD ERROR: Path not specified")  # DEBUG
			return False
		
		try:
			data = self.load_data(path)
		except:
			print("LOAD ERROR: %s" % (str(sys.exc_info())))  # DEBUG
			return False
		
		if "classes" in data:
			if not legacy_data_to_store(data, store, path, progress = self._progress):
				print("LOAD ERROR: Invalid legacy file format")  # DEBUG
				return False
		
		else:
			for name in ["object_relation_graph", "class_relation_graph", "class_membership_graph", 
				"local_folder", "max_order", "deposit_version", "user_tools", "queries",
			]:
				if name not in data:
					print("LOAD ERROR: Invalid file format")  # DEBUG
					return False
			
			if not self.data_to_store(data, store):
				print("LOAD ERROR: Loading data")  # DEBUG
				return False
		
		store.set_datasource(self)
		
		return True
	
	def data_to_store(self, data, store):
		# re-implement
		
		return False
	
	def __str__(self):
		
		return "%s (%s)" % (self.__class__.__name__, self.get_path())




import re
import ast

def get_new_key(prefix, substr, done):
	
	key = "_%s1_" % (prefix)
	n = 1
	while (key in done) or (key in substr):
		n += 1
		key = "_%s%d_" % (prefix, n)
	return key

def add_bracketed(clsname, descrname, prefix, substr, bracketed, lookup):
	
	if isinstance(clsname, str):
		clsname = clsname.strip("[]")
	
	if (clsname, descrname) in lookup:
		key = lookup[(clsname, descrname)]
	else:
		key = get_new_key(prefix, substr, bracketed)
		bracketed[key] = (clsname, descrname)
		lookup[(clsname, descrname)] = key
	
	return key

def str_to_key(substr):
	
	substr = tuple(substr.split("."))
	if len(substr) == 1:
		return (substr[0], None)
	return substr

def remove_quoted(substr):
	# returns substr, quoted = {key: quoted substr, ...}
	
	quoted = {}
	while True:
		found = False
		for m in re.finditer(r"(['\"])(.*?)\1", substr):
			i, j = m.start(0), m.end(0)
			key = get_new_key("QU", substr, quoted)
			quoted[key] = substr[i:j]
			substr = substr[:i] + " %s " % (key) + substr[j:]
			found = True
			break
		if not found:
			break
	
	return substr, quoted

def remove_bracketed_selects(substr, classes, descriptors):
	# returns substr, bracketed = {key: (class, descriptor), ...}
	
	def process_found(substr_cleaned, i, j, names):
		
		if substr_cleaned[:i].endswith("*."):
			names.append([i-2, i-1, "*", False])
			substr_cleaned = substr_cleaned[:i-2] + " " + substr_cleaned[i-1:]
		if substr_cleaned[j:].startswith(".*"):
			names.append([j+1, j+2, "*", False])
			substr_cleaned = substr_cleaned[:j+1] + " " + substr_cleaned[j+2:]
		
		return substr_cleaned
	
	bracketed = {}
	done = {}
	
	substr_cleaned = substr
	names = []  # [[i, j, text, is_bracketed], ...]
	
	for m in re.finditer(r"OBJ+\(.*?\)", substr_cleaned):
		i, j = m.start(0), m.end(0)
		ids = [id_.strip() for id_ in substr_cleaned[i+4:j-1].split(",")]
		if not False in [id_.isdigit() for id_ in ids]:
			names.append([i, j, tuple([int(id_) for id_ in ids]), True])
			substr_cleaned = substr_cleaned[:i] + (j-i)*" " + substr_cleaned[j:]
			substr_cleaned = process_found(substr_cleaned, i, j, names)
	
	for clsname in sorted(list(classes), key = lambda name: len(name))[::-1]:
		for m in re.finditer(r"(\A|[^a-zA-Z0-9_])(%s)([^a-zA-Z0-9_]|\Z)" % (clsname), substr_cleaned):
			i, j = m.start(0), m.end(0)
			found = False
			if (substr_cleaned[i] == "[") and (substr_cleaned[j-1] == "]"):
				names.append([i, j, substr_cleaned[i:j], True])
				substr_cleaned = substr_cleaned[:i] + (j-i)*" " + substr_cleaned[j:]
				found = True
			else:
				if substr_cleaned[i:].startswith(clsname):
					i -= 1
				if substr_cleaned[:j].endswith(clsname):
					j += 1
				if substr_cleaned[i+1:j-1].isidentifier():
					i += 1
					j -= 1
					names.append([i, j, substr_cleaned[i:j], False])
					substr_cleaned = substr_cleaned[:i] + len(substr_cleaned[i:j])*" " + substr_cleaned[j:]
					found = True
			if found:
				substr_cleaned = process_found(substr_cleaned, i, j, names)
	
	fragments = []  # [[text, is_class, is_bracketed], ...]
	j_last = 0
	for i, j, name, is_bracketed in sorted(names, key = lambda row: row[0]):
		if i > j_last:
			fragments.append([substr_cleaned[j_last:i], False, False])
		fragments.append([name, True, is_bracketed])
		j_last = j
	if j_last < len(substr_cleaned):
		fragments.append([substr_cleaned[j_last:], False, False])
	
	checkstr = ""
	for text, is_class, _ in fragments:
		if is_class:
			checkstr += "C"
		elif text == ".":
			checkstr += "."
		else:
			checkstr += "X"
	
	descriptors = descriptors.union({"*"})
	
	collect = []
	while fragments:
		if checkstr.startswith("C.C.C"):
			# not a SELECT (probably a relation)
			for i in range(5):
				text, _, _ = fragments.pop(0)
				collect.append(text)
			checkstr = checkstr[5:]
		elif checkstr.startswith("C.C"):
			clsname, descrname = fragments[0][0], fragments[2][0].strip("[]")
			clsname = clsname.strip("[]") if isinstance(clsname, str) else clsname
			if (descrname in descriptors) and (fragments[0][2] or fragments[2][2]):
				# second class is a descriptor and one of them is bracketed
				key = add_bracketed(clsname, descrname, "VAR", substr, bracketed, done)
				collect += " %s " % (key)
			else:
				collect += [text for text, _, _ in fragments[:3]]
			fragments = fragments[3:]
			checkstr = checkstr[3:]
		elif checkstr.startswith("CX") and fragments[1][0].strip().startswith("."):
			collect += [text for text, _, _ in fragments[:3]]
			fragments = fragments[3:]
			checkstr = checkstr[3:]
		elif checkstr.startswith("C") and fragments[0][2]:
			# bracketed class without a descriptor
			key = add_bracketed(fragments[0][0], None, "VAR", substr, bracketed, done)
			collect += " %s " % (key)
			fragments = fragments[1:]
			checkstr = checkstr[1:]
		else:
			text, _, _ = fragments.pop(0)
			collect.append(text)
			checkstr = checkstr[1:]
	collect = ["OBJ(%s)" % (",".join([str(id_) for id_ in item])) if isinstance(item, tuple) else item for item in collect]
	substr = "".join(collect)
	
	return substr, bracketed

def remove_classless(substr, classes):
	
	classless = set()
	while True:
		found = False
		for m in re.finditer("\!\*\.", substr):
			i, j = m.start(0), m.end(0)
			key = get_new_key("CL", substr, classless.union(classes))
			classless.add(key)
			substr = substr[:i] + key + substr[j-1:]
			found = True
			break
		if not found:
			break
	
	return substr, classless

def remove_bracketed_all(substr):
	# returns substr, bracketed = {key: bracketed substr, ...}
	
	bracketed = {}
	while True:
		found = False
		for m in re.finditer(r"OBJ+\(.*?\)", substr):
			i, j = m.start(0), m.end(0)
			ids = [id_.strip() for id_ in substr[i+4:j-1].split(",")]
			if not False in [id_.isdigit() for id_ in ids]:
				key = get_new_key("BA", substr, bracketed)
				bracketed[key] = tuple([int(id_) for id_ in ids])
				substr = substr[:i] + key + substr[j:]
				found = True
				break
		for m in re.finditer(r"\[.*?\]", substr):
			i, j = m.start(0), m.end(0)
			key = get_new_key("BA", substr, bracketed)
			bracketed[key] = substr[i:j]
			substr = substr[:i] + key + substr[j:]
			found = True
			break
		if not found:
			break
	
	return substr, bracketed

def replace_bracketed(substr, bracketed):
	
	for key in bracketed:
		data = bracketed[key]
		if isinstance(data, tuple):
			data = "OBJ(%s)" % (",".join([str(obj_id) for obj_id in data]))
		substr = substr.replace(key, data)
	return substr

class SelectsToVarsTransformer1(ast.NodeTransformer):
	
	def __init__(self, substr, classes, descriptors, bracketed_selects):
		# classes = {class_name, ...}
		# descriptors = {descriptor_name, ...}
		# bracketed_selects = {key: (class, descriptor), ...}
		
		ast.NodeTransformer.__init__(self)
		
		self._done = {}  # {(class_name, descriptor_name): variable_name, ...}
		self.selects = bracketed_selects  # {key: (class, descriptor), ...}
		
		self._substr = substr
		self._classes = classes
		self._descriptors = descriptors
		
		for name in self.selects:
			key = self.selects[name]
			self._done[key] = name
	
	def visit_Attribute(self, node):
		
		self.generic_visit(node)
		
		if (type(node.value) is ast.Name) and \
			(node.value.id not in __builtins__) and \
			(node.value.id in self._classes) and \
			(node.attr not in __builtins__) and \
			(node.attr in self._descriptors):
				name = add_bracketed(node.value.id, node.attr, "VAR", self._substr, self.selects, self._done)
				return ast.Name(id = name, ctx = node.ctx)
		
		return node

class SelectsToVarsTransformer2(SelectsToVarsTransformer1):
	
	def visit_Name(self, node):
		
		name = node.id
		if (name not in __builtins__) and (name in self._classes):
			name = add_bracketed(name, None, "VAR", self._substr, self.selects, self._done)
		
		return ast.Name(id = name, ctx = node.ctx)
	
	def visit_Attribute(self, node):
		
		return node

def extract_expr_vars(substr, classes, descriptors, bracketed_selects):
	# classes = {name, ...}
	# descriptors = {name, ...}
	# bracketed_selects = {key: (class, descriptor), ...}
	
	substr, classless = remove_classless(substr, classes)
	
	tree = ast.parse(substr)
	# find and replace Class.Descriptor occurences in substr
	transformer = SelectsToVarsTransformer1(substr, classes.union(classless), descriptors, bracketed_selects)
	tree = transformer.visit(tree)
	# find and replace Class occurences in substr
	transformer = SelectsToVarsTransformer2(substr, classes.union(classless), descriptors, bracketed_selects)
	tree = transformer.visit(tree)
	expr = ast.unparse(tree)
	vars = transformer.selects.copy()
	
	for name in vars:
		if vars[name][0] in classless:
			cls, descr = vars[name]
			vars[name] = ("!*", descr)
	
	return expr, vars


class Parse(object):
	
	KEYWORDS = ["SELECT", "COUNT", "SUM", "AS", "RELATED", "WHERE", "GROUP BY"]

	def __init__(self, querystr, classes, descriptors):
		# classes = {name, ...}
		# descriptors = {name, ...}
		# querystr = "SELECT [select1], [select2], COUNT([select1]) AS [alias], SUM([select1]) AS [alias], ... RELATED [relation1], [relation2], ... WHERE [conditions expression] GROUP BY [select1], [select2], ..."
		#	select = Class or Class.Descriptor
		#	relation = Class1.Relation.Class2
		#	conditions = python expression; e.g. Class.Descriptor > 3 or Class is not None
		#	use [name with spaces] to escape class, descriptor, relation or alias names containing spaces or KEYWORDS
		
		self.querystr = querystr
		self.classes = classes
		self.descriptors = descriptors
		
		self.columns = []  # [(class_name, descriptor_name), (None, alias), ...]
		self.selects = []  # [(class_name, descriptor_name), ...]
		self.group_by = []  # [(class_name, descriptor_name), ...]
		self.counts = []  # [(alias, class_name, descriptor_name), ...]
		self.sums = []  # [(alias, class_name, descriptor_name), ...]
		self.relations = []  # [(class1, relation, class2), ...]
		self.where_expr = ""
		self.where_vars = {}  # {name: (class, descriptor), ...}
		self.classes_used = []  # [name, ...]
		
		querystr, quoted = remove_quoted(self.querystr)
		querystr, bracketed_selects = remove_bracketed_selects(querystr, self.classes, self.descriptors)
		querystr, bracketed_other = remove_bracketed_all(querystr)
		
		collect = []
		for keyword in self.KEYWORDS:
			for m in re.finditer(r"(?i)\b%s\b" % (keyword), querystr):
				i, j = m.start(0), m.end(0)
				collect.append((i, j, keyword))
		collect = sorted(collect)
		collect2 = []
		for idx in range(len(collect)):
			i, j, keyword = collect[idx]
			if idx < len(collect) - 1:
				k = collect[idx + 1][0]
			else:
				k = len(querystr)
			collect2.append((keyword, j, k))
		collect = collect2
		collect2 = None
		
		while collect:
			keyword, i, j = collect.pop(0)
			substr = querystr[i:j].strip().strip(",").strip()
			
			if keyword in ["SELECT", "GROUP BY"]:
				for name in substr.split(","):
					name = name.strip()
					if name in bracketed_selects:
						item = bracketed_selects[name]
					else:
						item = str_to_key(name)
					if keyword == "SELECT":
						self.selects.append(item)
						self.columns.append(item)
					else:
						self.group_by.append(item)
			
			elif keyword in ["COUNT", "SUM"]:
				name = substr.strip("()").strip()
				if name in bracketed_selects:
					class_name, descriptor_name = bracketed_selects[name]
				else:
					class_name, descriptor_name = str_to_key(name)
				if collect and (collect[0][0] == "AS"):
					_, i, j = collect.pop(0)
					alias = querystr[i:j]
					alias = replace_bracketed(alias, bracketed_other)
					alias = alias.strip().strip(",").strip()
					class_name, descriptor_name, alias = [(item.strip("[]") if item else None) for item in [class_name, descriptor_name, alias]]
					if keyword == "COUNT":
						self.counts.append((alias, class_name, descriptor_name))
					else:
						self.sums.append((alias, class_name, descriptor_name))
					self.columns.append((None, alias))
			
			elif keyword == "RELATED":
				for chain in substr.split(","):
					chain = str_to_key(chain.strip())
					if len(chain) == 3:
						relation = []
						for item in chain:
							if item in bracketed_other:
								relation.append(bracketed_other[item])
								if isinstance(relation[-1], str):
									relation[-1] = relation[-1].strip("[]")
							else:
								relation.append(item.strip("[]"))
						self.relations.append(tuple(relation))
			
			elif keyword == "WHERE":
				substr = replace_bracketed(substr, bracketed_other)
				self.where_expr, self.where_vars = extract_expr_vars(substr, self.classes, self.descriptors, bracketed_selects)
		
		for key in quoted:
			self.where_expr = self.where_expr.replace(key, quoted[key])
		
		for class_name, _ in self.selects + self.group_by:
			if (class_name is not None) and (class_name not in self.classes_used):
				self.classes_used.append(class_name)
		for _, class_name, _ in self.counts + self.sums:
			if (class_name is not None) and (class_name not in self.classes_used):
				self.classes_used.append(class_name)
		for class_name1, _, class_name2 in self.relations:
			for class_name in [class_name1, class_name2]:
				if (class_name is not None) and (class_name not in self.classes_used):
					self.classes_used.append(class_name)
		for name in self.where_vars:
			class_name = self.where_vars[name][0]
			if class_name not in self.classes_used:
				self.classes_used.append(class_name)
			



from deposit.store.abstract_dtype import AbstractDType
from deposit.utils.fnc_geometry import (coords_to_wkt, wkt_to_coords)

class DGeometry(AbstractDType):
	
	@property
	def geometry_type(self):
		
		return self.get_value()[0]
	
	@property
	def coords(self):
		
		return self.get_value()[1]
	
	@property
	def srid(self):
		
		return self.get_value()[2]
	
	@property
	def srid_vertical(self):
		
		return self.get_value()[3]
	
	@property
	def wkt(self):
		
		value = AbstractDType.get_value(self)
		if isinstance(value, tuple):
			return coords_to_wkt(*value)
		return value
	
	def get_value(self):
		
		value = AbstractDType.get_value(self)
		if value is None:
			AbstractDType.set_value(self, (None, None, None, None))
		elif isinstance(value, str):
			AbstractDType.set_value(self, wkt_to_coords(value))
		return AbstractDType.get_value(self)
	
	def set_value(self, value):
		# value = DGeometry or 
		#	WKT or 
		# 	(geometry_type, coords, srid, srid_vertical) or
		#	(geometry_type, coords, srid) or
		#	(geometry_type, coords)
		
		if isinstance(value, DGeometry):
			AbstractDType.set_value(self, value.value)
			return
		
		if (value is None) or isinstance(value, str):
			AbstractDType.set_value(self, value)
			return
		
		if (not (isinstance(value, tuple) or isinstance(value, list))) or (len(value) < 2):
			raise Exception("Invalid value specified: %s" % (str(value)))
		args = [None, None, -1, -1]
		args[:len(value)] = value
		if isinstance(args[1], tuple):
			args[1] = list(args[1])
		if not isinstance(args[1], list):
			raise Exception("Invalid coords specified: %s" % (str(args[1])))
		value = tuple(args)
		AbstractDType.set_value(self, value)
	
	def to_dict(self):
		
		value = AbstractDType.get_value(self)
		if isinstance(value, tuple):
			value = coords_to_wkt(*value)
		
		return dict(
			dtype = self.__class__.__name__,
			value = value,
		)
	
	def __str__(self):
		
		return "%s(%s)" % (self.__class__.__name__, self.geometry_type)



import json

def add_srid_to_wkt(wkt, srid, srid_vertical):
	
	if (srid_vertical is not None) and (srid_vertical > -1):
		wkt = ("VERT_SRID=%d;" % (srid_vertical)) + wkt
	if (srid is not None) and (srid > -1):
		wkt = ("SRID=%d;" % (srid)) + wkt
	return wkt

def split_srid_from_wkt(wkt):
	
	def _extract_srid(substr, prefix):
		
		srid = None
		if substr.startswith(prefix):
			i = substr.find(";")
			if i == -1:
				raise Exception("Error in WKT: %s" % (substr))
			srid = int(substr[:i-1].split("=")[1].strip())
			substr = substr[i+1:]
		return srid, substr
	
	srid, wkt = _extract_srid(wkt.strip(), "SRID")
	srid_vertical, wkt = _extract_srid(wkt, "VERT_SRID")
	
	return wkt, srid, srid_vertical

def coords_to_wkt(geom_type: str, coords: list, srid: int = None, srid_vertical: int = None) -> str:
	
	def _to_list(crds):
		
		for i in range(len(crds)):
			if isinstance(crds[i], tuple):
				crds[i] = list(crds[i])
			if isinstance(crds[i], list):
				_to_list(crds[i])
	
	def _close_polygons(crds):
		
		if not isinstance(crds[0], list):
			return
		if isinstance(crds[0][0], list):
			for crd in crds:
				_close_polygons(crd)
		elif crds[0] != crds[-1]:
			crds.append(crds[0])
	
	def _points_to_wkt(points):
		
		return "(%s)" % (", ".join([" ".join([str(val) for val in point]) for point in points]))
	
	def _polygon_to_wkt(polygon):
		
		return "(%s)" % ", ".join([_points_to_wkt(points) for points in polygon])
	
	if not coords:
		return None
	
	_to_list(coords)
	
	geom_type = geom_type.upper()
	is_z = geom_type.endswith("Z")
	is_m = geom_type.endswith("M")
	if is_z:
		geom_type = geom_type.rstrip("Z")
	if is_m:
		geom_type = geom_type.rstrip("M")
	ndims = 1
	ncoords = len(coords)
	c = coords[0]
	while isinstance(c, list):
		ncoords = len(c)
		c = c[0]
		ndims += 1
	if is_z and ncoords != 3:
		raise Exception("Invalid number of coordinates (%d). Required: 3." % (ncoords))
	if is_m and ncoords != 4:
		raise Exception("Invalid number of coordinates (%d). Required: 4." % (ncoords))
	is_z = (ncoords == 3)
	is_m = (ncoords == 4)
	
	ndims_required = {
		"POINT": 1,
		"MULTIPOINT": 2,
		"LINESTRING": 2,
		"POLYGON": 3,
		"MULTIPOLYGON": 4,
	}[geom_type]
	if ndims != ndims_required:
		raise Exception("Invalid number of dimensions (%d) for geometry type %s. Required: %d." % (ndims, geom_type, ndims_required))
	
	if geom_type.endswith("POLYGON"):
		_close_polygons(coords)
	
	if is_z:
		geom_type += "Z"
	elif is_m:
		geom_type += "M"
	
	wkt = geom_type
	if ndims == 1:
		wkt += "(%s)" % (" ".join([str(val) for val in coords]))
	elif ndims == 2:
		wkt += _points_to_wkt(coords)
	elif ndims == 3:
		wkt += _polygon_to_wkt(coords)
	elif ndims == 4:
		wkt += "(%s)" % ", ".join([_polygon_to_wkt(polygon) for polygon in coords])
	
	wkt = add_srid_to_wkt(wkt, srid, srid_vertical)
	
	return wkt

def wkt_to_coords(wkt):
	# returns (geometry_type, coordinates, srid, srid_vertical)
	
	def _extract_srid(substr, prefix):
		
		srid = None
		if substr.startswith(prefix):
			i = substr.find(";")
			if i == -1:
				raise Exception("Error in WKT: %s" % (substr))
			srid = int(substr[:i-1].split("=")[1].strip())
			substr = substr[i+1:]
		return srid, substr
	
	wkt = wkt.strip()
	wkt = wkt.replace("(","[")
	wkt = wkt.replace(")","]")
	
	wkt, srid, srid_vertical = split_srid_from_wkt(wkt)
	
	i = wkt.find("[")
	geometry_type = wkt[:i].strip().upper()
	wkt = wkt[i:]
	
	collect = ""
	while True:
		j = wkt.find(",")
		if j == -1:
			j = wkt.find("]")
		if j > -1:
			substr = wkt[:j]
			i = substr.find("[")
			collect += wkt[:i+1]
			substr = ",".join([item.strip() for item in wkt[i+1:j].strip().split(" ")])
			if substr:
				collect += "[%s]" % (substr)
			collect += wkt[j]
			wkt = wkt[j+1:]
		else:
			break
	if collect:
		wkt = collect + wkt
	coords = json.loads(wkt)
	
	if geometry_type.startswith("POINT"):
		coords = coords[0]
	
	return geometry_type, coords, srid, srid_vertical





class AbstractDatasource(object):
	
	def __init__(self):
		
		self._progress = None
	
	def to_dict(self):
		
		return dict(
			datasource = self.__class__.__name__,
			url = None,
			identifier = None,
			connstr = None,
		)
	
	def get_name(self):
		
		return "data"
	
	def set_progress(self, progress):
		
		self._progress = progress
	
	def close_progress(self):
		
		if self._progress is not None:
			self._progress.cancel()
	
	def update_progress(self, value, maximum = None, text = None):
		
		if self._progress is not None:
			self._progress.update_state(text = text, value = value, maximum = maximum)
	
	def progress_cancelled(self):
		
		if self._progress is not None:
			return self._progress.cancel_pressed()
		return False
	
	def is_valid(self):
		
		return False
	
	def can_create(self):
		
		return False
	
	def get_folder(self):
		
		return None
	
	def create(self, *args, **kwargs):
		
		pass
	
	def save(self, store, progress = None, *args, **kwargs):
		
		pass
	
	def load(self, store, progress = None, *args, **kwargs):
		
		pass
	
	def __str__(self):
		
		return self.__class__.__name__



from deposit.externalsource.abstract_externalsource import AbstractExternalsource
from deposit.store.dgeometry import DGeometry
from deposit.utils.fnc_geometry import (add_srid_to_wkt, split_srid_from_wkt)
from deposit.utils.fnc_serialize import (try_numeric)

from shapefile import (Reader, Writer)
from shapely.geometry import shape as shapely_shape
from shapely import wkt as shapely_wkt

class SHP(AbstractExternalsource):
	
	def __init__(self):
		
		AbstractExternalsource.__init__(self)
		
		self._columns = {} # {column_idx: name, ...}
		self._data = [] # [{column_idx: value, ...}, ...]
	
	def sheets(self):
		
		return ["0"]
	
	def column_count(self, sheet):
		
		return len(self._columns)
	
	def row_count(self, sheet):
		
		return len(self._data)
	
	def column_name(self, sheet, column_idx):
		
		if column_idx not in self._columns:
			return None
		return self._columns[column_idx]
	
	def data(self, sheet, row_idx, column_idx):
		
		if row_idx > len(self._data) - 1:
			return None
		if column_idx not in self._data[row_idx]:
			return None
		return self._data[row_idx][column_idx]
	
	def save(self, get_header, get_item, n_rows, n_cols, path = None, url = None, *args, **kwargs):
		# get_header(col, user_role = False) = column_name or (class_name, descriptor_name) if user_role = True
		# get_item(row, col) = deposit_gui.QueryItem
		
		def abbrev_to(name, chars, columns_abbrev):
			
			if len(name) > chars:
				n = 1
				while True:
					name_new = name[:chars - len(str(n))] + str(n)
					if not name_new in columns_abbrev.values():
						return name_new
					n += 1
			return name
		
		self.set_path(path, url)
		path = self.get_path()
		if path is None:
			return False
		
		geometries = []  # [(Shapely geometry, geometry type), ...]
		row_idxs = [] # [row, ...]
		for row in range(n_rows):
			for col in range(n_cols):
				item = get_item(row, col)
				if item.is_geometry():
					wkt, srid, srid_vertical = split_srid_from_wkt(item.value.wkt)
					# TODO implement saving srid, srid_vertical
					geometries.append((shapely_wkt.loads(wkt), item.value.geometry_type))
					row_idxs.append(row)
					break
		
		if not row_idxs:
			return False
		
		column_names = self.get_header_data(get_header, n_cols)
		columns_abbrev = {} # {col: column_abbrev, ...}; abbreviated column names
		for col in range(n_cols):
			column_abbrev = column_names[col]
			if len(column_abbrev) > 10:
				class_name, descriptor_name = get_header(col, user_role = True)
				names = [name for name in [class_name, descriptor_name] if isinstance(name, str)]
				if len(names) == 2:
					column_abbrev = "_".join([abbrev_to(names[0], 4, columns_abbrev), abbrev_to(names[1], 5, columns_abbrev)])
				elif len(names) == 1:
					column_abbrev = abbrev_to(names[0], 10, columns_abbrev)
				else:
					column_abbrev = abbrev_to("no class", 10, columns_abbrev)
			column_abbrev = column_abbrev.replace(".", "_")
			columns_abbrev[col] = column_abbrev
		
		shapeType = -1
		shape_types = {
			"POINT": 1,
			"LINESTRING": 3,
			"POLYGON": 5,
			"MULTIPOINT": 8,
			"POINTZ": 11,
			"LINESTRINGZ": 13,
			"POLYGONZ": 15,
			"MULTIPOINTZ": 18,
			"POINTM": 21,
			"LINESTRINGM": 23,
			"POLYGONM": 25,
			"MULTIPOINTM": 28,
		}
		for _, geometry_type in geometries:
			if geometry_type not in shape_types:
				raise Exception("Unknown geometry type")
			if shapeType > -1:
				if shape_types[geometry_type] != shapeType:
					raise Exception("Geometries must be of the same type")
			else:
				shapeType = shape_types[geometry_type]
		
		sf = Writer(path, shapeType = shapeType)
		types = {} # {col: type, ...}
		shp_types = {bool: "C", int: "N", float: "N", str: "C"}
		conv_order = ["N", "C"]
		data = {}
		for i, row in enumerate(row_idxs):
			data[i] = {}
			for col in range(n_cols):
				item = get_item(row, col)
				if item.is_geometry():
					continue
				data[i][col] = try_numeric(item.get_display_data())
				typ = type(data[i][col])
				typ = shp_types[typ] if typ in shp_types else "C"
				if (not col in types) or (
					(typ != types[col]) and \
					(conv_order.index(typ) > conv_order.index(types[col]))
				):
					types[col] = typ
		for col in range(n_cols):
			if col in types:
				sf.field(columns_abbrev[col], fieldType = types[col], size = "128")
		for i in range(len(geometries)):
			sf.shape(geometries[i][0].__geo_interface__)
			if types:
				sf.record(*[data[i].get(col,None) for col in range(n_cols) if col in types])
		sf.close()
		
		return True
	
	def load(self, path = None, url = None, *args, **kwargs):
		
		self.set_path(path, url)
		path = self.get_path()
		if path is None:
			return False
		
		self._columns.clear()
		self._data.clear()
		
		# get SRID
		srid = -1
		srid_vertical = -1
		# TODO try to discover SRID from .prj file
		
		# get geometries & data
		geo_column = "Geometry"
		sf = Reader(path)
		names = []
		for idx, name in enumerate(sf.fields[1:]):
			self._columns[idx + 1] = name[0]
			names.append(name[0])
		while geo_column in names:
			geo_column = geo_column + "_"
		
		geometries = [add_srid_to_wkt(shapely_shape(shape.__geo_interface__).wkt, srid, srid_vertical) for shape in sf.shapes()]
		# geometries = [wkt definition, ...] in order of records
		for i, record in enumerate(sf.records()):
			self._data.append(dict([(idx, str(record[idx - 1]).strip()) for idx in self._columns]))
			self._data[-1][0] = DGeometry((geometries[i]))
		self._columns[0] = geo_column
		
		return True




from deposit.externalsource.abstract_externalsource import AbstractExternalsource

import csv

class CSV(AbstractExternalsource):
	
	def __init__(self):
		
		AbstractExternalsource.__init__(self)
		
		self._columns = {} # {column_idx: name, ...}
		self._data = [] # [{column_idx: value, ...}, ...]
	
	def sheets(self):
		
		return ["0"]
	
	def column_count(self, sheet):
		
		return len(self._columns)
	
	def row_count(self, sheet):
		
		return len(self._data)
	
	def column_name(self, sheet, column_idx):
		
		if column_idx not in self._columns:
			return None
		return self._columns[column_idx]
	
	def data(self, sheet, row_idx, column_idx):
		
		if row_idx > len(self._data) - 1:
			return None
		if column_idx not in self._data[row_idx]:
			return None
		return self._data[row_idx][column_idx]
	
	def save(self, get_header, get_item, n_rows, n_cols, path = None, url = None, *args, **kwargs):
		# get_header(col, user_role = False) = column_name or (class_name, descriptor_name) if user_role = True
		# get_item(row, col) = deposit_gui.QueryItem
		
		self.set_path(path, url)
		path = self.get_path()
		if path is None:
			return False
		
		with open(path, "w", newline = "") as f:
			writer = csv.writer(f, dialect=csv.excel, quoting=csv.QUOTE_ALL)
			writer.writerow(self.get_header_data(get_header, n_cols))
			for row in range(n_rows):
				writer.writerow(self.get_row_data(row, get_item, n_cols))
		
		return True
	
	def load(self, path = None, url = None, *args, **kwargs):
		
		self.set_path(path, url)
		path = self.get_path()
		if path is None:
			return False
		
		self._columns.clear()
		self._data.clear()
		
		with open(path, "r", newline = "") as f:
			reader = csv.reader(f, dialect = csv.excel, quoting=csv.QUOTE_ALL)
			idx_lookup = {} # {i: idx, ...}
			for row in reader:
				if not self._columns:
					idx = 0
					for i, value in enumerate(row):
						value = value.strip()
						if value:
							self._columns[idx] = value
							idx_lookup[i] = idx
							idx += 1
					if not self._columns:
						return True
				else:
					self._data.append(dict([(idx_lookup[i], value.strip()) for i, value in enumerate(row) if i in idx_lookup]))
		return True




from deposit import __version__
from deposit.datasource.abstract_datasource import AbstractDatasource
from deposit.store.abstract_delement import AbstractDElement

from deposit.utils.fnc_serialize import (parse_connstr, legacy_data_to_store, json_data_to_store, GRAPH_ATTRS)

from collections import defaultdict
from natsort import natsorted
import psycopg2
import sys

class AbstractDBSource(AbstractDatasource):
	
	def __init__(self):
		
		AbstractDatasource.__init__(self)
		
		self._username = None
		self._password = None
		self._host = None
		self._database = None
		self._schema = None
		self._identifier = None
		
		self._cursor = None
		self._tables = []  # [tablename, ...]
		self._schemas = []  # [schemaname, ...]
	
	def get_name(self):
		
		if self._identifier is not None:
			return self._identifier
		return "data"
	
	def to_dict(self):
		
		data = AbstractDatasource.to_dict(self)
		data["identifier"] = self.get_identifier()
		data["connstr"] = self.get_connstr()
		
		return data
	
	def get_username(self):
		
		return self._username
	
	def get_password(self):
		
		return self._password
	
	def get_host(self):
		
		return self._host
	
	def get_database(self):
		
		return self._database
	
	def get_schema(self):
		
		return self._schema
	
	def get_identifier(self):
		
		return self._identifier
	
	def get_connstr(self):
		
		username = self.get_username()
		password = self.get_password()
		host = self.get_host()
		database = self.get_database()
		schema = self.get_schema()
		if None in [username, password, host, database, schema]:
			return None
		
		return "postgres://%s:%s@%s/%s?currentSchema=%s" % (username, password, host, database, schema)
	
	def get_cursor(self):
		
		if self._cursor is None:
			self.connect()
		return self._cursor
	
	def get_schemas(self):
		
		return self._schemas
	
	def get_tables(self):
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		if (not self._tables) and (cursor):
			schema = self.get_schema()
			if schema in self._schemas:
				cursor.execute("SELECT tablename FROM pg_tables WHERE schemaname = '%s';" % (schema))
				self._tables = natsorted([row[0] for row in cursor.fetchall()])
		if not was_connected:
			self.disconnect()
		return self._tables
	
	def set_username(self, value):
		
		self._username = value
	
	def set_password(self, value):
		
		self._password = value
	
	def set_host(self, value):
		
		self._host = value
	
	def set_database(self, value):
		
		self._database = value
	
	def set_schema(self, value):
		
		self._schema = value
	
	def set_identifier(self, value):
		
		self._identifier = value
	
	def set_connstr(self, value):
		
		parsed = parse_connstr(value)
		if not parsed:
			return
		self.set_username(parsed["username"])
		self.set_password(parsed["password"])
		self.set_host(parsed["host"])
		self.set_database(parsed["dbname"])
		self.set_schema(parsed["schema"])
	
	def connect(self):
		
		if self._cursor is not None:
			self._cursor.connection.close()
		self._cursor = None
		self._tables = []
		self._schemas = []
		
		dbname = self.get_database()
		user = self.get_username()
		password = self.get_password()
		host = self.get_host()
		port = 5432
		if None in [dbname, user, password, host]:
			return False
		if ":" in host:
			host, port = host.split(":")
			if not port.isdigit():
				return False
			port = int(port)
		
		try:
			conn = psycopg2.connect(
				dbname = dbname,
				user = user,
				password = password,
				host = host,
				port = port,
			)
		except:
			_, exc_value, _ = sys.exc_info()
			print("Connection Error: %s" % (str(exc_value)))
			return False
		self._cursor = conn.cursor()
		if self._cursor is not None:
			self._cursor.execute("SELECT schema_name FROM information_schema.schemata")
			self._schemas = natsorted([row[0] for row in self._cursor.fetchall()])
		
		return True
	
	def disconnect(self):
		
		if self._cursor is not None:
			conn = self._cursor.connection
			self._cursor.close()
			conn.close()
			self._cursor = None
	
	
	def is_connected(self):
		
		if (self._cursor is None) or self._cursor.closed:
			return False
		return True
	
	def is_valid(self):
		
		return False
	
	def can_create(self):
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		ret = (cursor is not None)
		
		if not was_connected:
			self.disconnect()
		return ret
	
	def get_folder(self):
		
		return None
	
	def create(self):
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		schema = self.get_schema()
		if schema:
			cursor.execute("CREATE SCHEMA IF NOT EXISTS %s" % (schema))
			cursor.connection.commit()
			ret = True
		else:
			ret = False
		
		if not was_connected:
			self.disconnect()
		
		return ret
	
	def delete(self):
		
		return False
	
	def save_data(self, 
		db_meta,
		object_nodes, object_relations, 
		class_nodes, class_relations, 
		member_nodes, member_relations, 
		user_tools, queries, 
		progress = None
	):
		# store data in postgres db
		'''
		db_meta = dict(
			deposit_version,
			local_folder,
			max_order,
		)
		object_nodes = [{id: int, data: object_node_data}, ...]
			object_node_data = {
				(optional) descriptors: {descr_name: value or AbstractDType.to_dict(), ...}
				(optional) locations: {descr_name: DGeometry.to_dict(), ...}
			}
		object_relations = [{src: int, tgt: int, lbl: str, weight: float}, ...]
		class_nodes = [{id: str, data: class_node_data}, ...]
			class_node_data = {
				order: int,
				(optional) descriptors = [descr_name, ...]
			}
		class_relations = [{src: str, tgt: str, lbl: str}, ...]
		member_nodes = [{id: int/str}, ...]
		member_relations = [{src: int/str, tgt: int/str}, ...]
		user_tools = [dict(), ...]
		queries = {name: querystr, ...}
		'''
		
		pass
	
	def save(self, store, progress = None, identifier = None, connstr = None, *args, **kwargs):
		
		def delements_to_dict(data):
			
			for node_data in data["nodes"]:
				if isinstance(node_data["data"], AbstractDElement):
					node_data["data"] = node_data["data"].to_dict()
		
		self.set_progress(progress)
		
		if not store.has_local_folder():
			return False
		
		if identifier is not None:
			self.set_identifier(identifier)
		if connstr is not None:
			self.set_connstr(connstr)
		if not self.can_create():
			return False
		self.create()
		
		db_meta = dict(
			deposit_version = __version__,
			local_folder = store._local_folder,
			max_order = store._max_order,
		)
		
		data = store.G.objects_to_json(GRAPH_ATTRS)
		delements_to_dict(data)
		object_nodes = data["nodes"]  # [{id: int, data: dict()}, ...]
		object_relations = data[GRAPH_ATTRS["link"]]  # [{src: int, tgt: int, lbl: str}, ...]
		
		data = store.G.classes_to_json(GRAPH_ATTRS)
		delements_to_dict(data)
		class_nodes = data["nodes"]  # [{id: str, data: dict()}, ...]
		class_relations = data[GRAPH_ATTRS["link"]]  # [{src: str, tgt: str, lbl: str}, ...]
		
		data = store.G.members_to_json(GRAPH_ATTRS)
		member_nodes = data["nodes"]  # [{id: int/str}, ...]
		member_relations = data[GRAPH_ATTRS["link"]]  # [{src: int/str, tgt: int/str}, ...]
		
		user_tools = store._user_tools  # [dict(), ...]
		
		queries = store._queries  # {name: querystr, ...}
		
		return self.save_data(
			db_meta,
			object_nodes, object_relations, 
			class_nodes, class_relations, 
			member_nodes, member_relations, 
			user_tools, queries, 
			progress
		)
	
	def load_data(self, progress = None):
		# returns data or legacy_data
		# data = dict(
		#	object_relation_graph = nx.node_link_data
		#	class_relation_graph = nx.node_link_data
		#	class_membership_graph = nx.node_link_data
		#	local_folder = str
		#	max_order = int
		#	user_tools = list
		#	queries = dict()
		#	deposit_version = str
		# )
		#
		# legacy_data = dict(
		#	classes = dict()
		#	objects = dict()
		#	user_tools = list
		#	queries = dict()
		#	local_folder = str
		# )
		
		return {}
		
	def load(self, store, progress = None, identifier = None, connstr = None, *args, **kwargs):
		
		self.set_progress(progress)
		
		if identifier is not None:
			self.set_identifier(identifier)
		if connstr is not None:
			self.set_connstr(connstr)
		if not self.is_valid():
			return False
		
		data = self.load_data(progress)  # DEBUG
#		try:
#			data = self.load_data(progress)
#		except:
#			print("LOAD ERROR: %s" % (str(sys.exc_info())))  # DEBUG
#			return False
		
		if "classes" in data:
			local_folder = data["local_folder"]
			if not legacy_data_to_store(data, store, local_folder, progress = self._progress):
				print("LOAD ERROR: Invalid legacy file format")  # DEBUG
				return False
		else:	
			if not json_data_to_store(data, store, self._progress):
				print("LOAD ERROR: Loading data")  # DEBUG
				return False
		
		store.set_datasource(self)
		
		return True
	
	def __str__(self):
		
		return "%s (%s)" % (self.__class__.__name__, str(self.get_identifier()))




class AbstractDType(object):
	
	def __init__(self, value = None, *args, **kwargs):
		
		self.set_value(value)
	
	@property
	def value(self):
		
		return self.get_value()
	
	@value.setter
	def value(self, value):
		
		self.set_value(value)
	
	def get_value(self):
		
		return self._value
	
	def set_value(self, value):
		
		self._value = value
	
	def to_dict(self):
		
		return dict(
			dtype = self.__class__.__name__,
			value = self._value,
		)
	
	def from_dict(self, data):
		
		self._value = data["value"]
		
		return self
		
	def __str__(self):
		
		return "%s(%s)" % (self.__class__.__name__, str(self.get_value()))
	
	def __repr__(self):
	
		return self.__str__()
	
	def __eq__(self, other):
		
		return (self.__class__ == other.__class__) and (self.get_value() == other.get_value())




from deposit.store.abstract_dtype import AbstractDType
from deposit.store.ddatetime import DDateTime
from deposit.store.dgeometry import DGeometry
from deposit.store.dresource import DResource
from deposit.store.ddict import DDict
from deposit.store.dobject import DObject
from deposit.store.dclass import DClass

from deposit.utils.fnc_files import (as_url, get_updated_local_url)
from deposit.utils.fnc_geometry import (add_srid_to_wkt)
from deposit.query.parse import (remove_quoted, remove_bracketed_all)

from collections import defaultdict
import datetime
import json
import re
import os

GRAPH_ATTRS = dict(
	source = "src", 
	target = "tgt", 
	name = "id",
	key = "lbl", 
	link = "relations"
)

def try_numeric(value):
	
	if isinstance(value, str):
		if value.isdigit():
			return int(value)
		if value.replace(".","",1).isdigit():
			return float(value)
	return value

def dtype_to_dict(value):
	
	if isinstance(value, AbstractDType):
		return value.to_dict()
	if isinstance(value, dict):
		return DDict(value).to_dict()
	return value

def value_to_dtype(value):
	
	if not isinstance(value, dict):
		return value
	value = globals()[value["dtype"]]().from_dict(value)
	if isinstance(value, DDict):
		return value.value
	return value

def value_to_str(value):
	
	if value is None:
		return None
	if isinstance(value, DDateTime):
		return value.isoformat
	if isinstance(value, DGeometry):
		return value.wkt
	if isinstance(value, DResource):
		return value.url
	return str(value)

def parse_connstr(connstr):
	
	values = re.match('postgres://(.*?):(.*?)@(.*?)/(.*)\?currentSchema=(.*)', connstr)
	if values is None:
		return {}
	username, password, host, dbname, schema = values.groups()
	return dict(
		username = username,
		password = password,
		host = host,
		dbname = dbname,
		schema = schema,
	)

def legacy_data_to_store(data, store, path, progress = None):
	
	for name in ["classes", "objects"]:
		if name not in data:
			return False
	
	store.clear()
	
	local_folder = None
	if "local_folder" in data:
		local_folder = data["local_folder"]
		store.set_local_folder(local_folder)
	if local_folder is None:	
		local_folder = os.path.normpath(os.path.abspath(os.path.dirname(path)))
	
	for obj_id in data["objects"]:
		obj = DObject(store, int(obj_id) + 1)
		store.G.add_object(int(obj_id) + 1, obj)
	
	for name in data["classes"]:
		order = data["classes"][name]["order"]
		cls = DClass(store, name, order)
		store.G.add_class(name, cls)
		store._max_order = max(store._max_order, order)
	
	if progress is not None:
		cmax = len(data["objects"]) + len(data["classes"]) + store.G.n_objects() + store.G.n_classes()
		cnt = 1
	
	class_lookup = defaultdict(set)  # {obj_id: set(class_name, ...), ...}
	
	cls_rels_done = set()  # [(src_class, tgt_class, label), ...]
	for name in data["classes"]:
		
		if progress is not None:
			if progress.cancel_pressed():
				store.clear()
				return True
			progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
		
		store.G.get_class_data(name).from_dict_1(
			dict(
				descriptors = data["classes"][name]["descriptors"] if "descriptors" in data["classes"][name] else {}
			)
		)
		
		for obj_id in data["classes"][name]["objects"]:
			if obj_id not in data["objects"]:
				continue
			store.G.add_class_child(name, int(obj_id) + 1)
			class_lookup[obj_id].add(name)
		
		for name_subclass in data["classes"][name]["subclasses"]:
			if name_subclass not in data["classes"]:
				continue
			store.G.add_class_child(name, name_subclass)
		
		if "relations" in data["classes"][name]:
			for label in data["classes"][name]["relations"]:
				for name_tgt in data["classes"][name]["relations"][label]:
					if name_tgt not in data["classes"]:
						continue
					store.G.add_class_relation(name, name_tgt, label)
					cls_rels_done.add((name, name_tgt, label))
	
	collect_rels = set()  # [(src_class, tgt_class, label), ...]
	for obj_id in data["objects"]:
		
		if progress is not None:
			if progress.cancel_pressed():
				store.clear()
				return True
			progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
		
		obj_data = dict(
			id = obj_id,
			descriptors = {},
			locations = {},
		)
		
		for name in data["objects"][obj_id]["descriptors"]:
			
			dtype = data["objects"][obj_id]["descriptors"][name]["label"]["dtype"]
			value = None
			
			if dtype == "DString":
				value = try_numeric(data["objects"][obj_id]["descriptors"][name]["label"]["value"])
			
			elif dtype == "DDateTime":
				try:
					value = datetime.datetime.fromisoformat(data["objects"][obj_id]["descriptors"][name]["label"]["value"])
				except:
					value = None
				if value is not None:
					value = dict(
						dtype = "DDateTime",
						value = value,
					)
			
			elif dtype == "DGeometry":
				wkt = data["objects"][obj_id]["descriptors"][name]["label"]["value"]
				srid = data["objects"][obj_id]["descriptors"][name]["label"]["srid"]
				srid_vertical = data["objects"][obj_id]["descriptors"][name]["label"]["srid_vertical"]
				wkt = add_srid_to_wkt(wkt, srid, srid_vertical)
				value = dict(
					dtype = "DGeometry",
					value = wkt,
				)
			
			elif dtype == "DResource":
				url = data["objects"][obj_id]["descriptors"][name]["label"]["value"]
				path = data["objects"][obj_id]["descriptors"][name]["label"]["path"]
				filename = data["objects"][obj_id]["descriptors"][name]["label"]["filename"]
				is_image = data["objects"][obj_id]["descriptors"][name]["label"]["image"]
				is_stored = (path is not None)
				if is_stored:
					url = get_updated_local_url(as_url(path), local_folder)
				if url is not None:
					value = dict(
						dtype = "DResource",
						value = (url, filename, is_stored, is_image), 
					)
			
			if value is not None:
				obj_data["descriptors"][name] = value
		
		store.G.get_object_data(int(obj_id) + 1).from_dict_1(obj_data)
		
		for label in data["objects"][obj_id]["relations"]:
			for obj_id_tgt in data["objects"][obj_id]["relations"][label]["objects"]:
				if obj_id_tgt not in data["objects"]:
					continue
				weight = None
				if ("weights" in data["objects"][obj_id]["relations"][label]) and \
					(obj_id_tgt in data["objects"][obj_id]["relations"][label]["weights"]):
					weight = data["objects"][obj_id]["relations"][label]["weights"][obj_id_tgt]
				store.G.add_object_relation(int(obj_id) + 1, int(obj_id_tgt) + 1, label, weight)
				for src_class in class_lookup[obj_id]:
					for tgt_class in class_lookup[obj_id_tgt]:
						collect_rels.add((src_class, tgt_class, label))
	collect_rels = cls_rels_done.difference(cls_rels_done)
	for src_class, tgt_class, label in collect_rels:
		store.G.add_class_relation(src_class, tgt_class, label)
	
	for obj in store.G.iter_objects_data():
		if progress is not None:
			if progress.cancel_pressed():
				store.clear()
				return True
			progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
		obj.from_dict_2()
	for cls in store.G.iter_classes_data():
		if progress is not None:
			if progress.cancel_pressed():
				store.clear()
				return True
			progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
		cls.from_dict_2()
	
	if "user_tools" in data:
		store._user_tools = data["user_tools"]
	
	if "queries" in data:
		store._queries = data["queries"]
	
	return True

def json_data_to_store(data, store, progress = None):
	
	n_nodes = len(data["object_relation_graph"]["nodes"]) + len(data["class_relation_graph"]["nodes"])
	cmax = 4*n_nodes
	cnt = 1
	if progress is not None:
		progress.update_state(value = cnt, maximum = cmax)
	
	for node in data["object_relation_graph"]["nodes"]:
		if progress is not None:
			if progress.cancel_pressed():
				return True
			progress.update_state(value = cnt, maximum = cmax)
		cnt += 1
		node["data"] = DObject(store, node["id"]).from_dict_1(node["data"])
	for node in data["class_relation_graph"]["nodes"]:
		if progress is not None:
			if progress.cancel_pressed():
				return True
			progress.update_state(value = cnt, maximum = cmax)
		cnt += 1
		node["data"] = DClass(store, node["id"], node["data"]["order"]).from_dict_1(node["data"])
	
	store.clear()
	store.G.objects_from_json(data["object_relation_graph"], GRAPH_ATTRS)
	if progress is not None:
		if progress.cancel_pressed():
			store.clear()
			return True
		progress.update_state(value = cnt, maximum = cmax)
	cnt += n_nodes / 3
	store.G.classes_from_json(data["class_relation_graph"], GRAPH_ATTRS)
	if progress is not None:
		if progress.cancel_pressed():
			store.clear()
			return True
		progress.update_state(value = cnt, maximum = cmax)
	cnt += n_nodes / 3
	store.G.members_from_json(data["class_membership_graph"], GRAPH_ATTRS)
	if progress is not None:
		if progress.cancel_pressed():
			store.clear()
			return True
		progress.update_state(value = cnt, maximum = cmax)
	cnt += n_nodes / 3
	for obj in store.G.iter_objects_data():
		if progress is not None:
			if progress.cancel_pressed():
				store.clear()
				return True
			progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
		obj.from_dict_2()
	for cls in store.G.iter_classes_data():
		if progress is not None:
			if progress.cancel_pressed():
				store.clear()
				return True
			progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
		cls.from_dict_2()
	store._max_order = data["max_order"]
	store._user_tools = data["user_tools"]
	store._queries = data["queries"]
	store._local_folder = data["local_folder"]
	
	if progress is not None:
		progress.update_state(value = cmax, maximum = cmax)
	
	return True

def select_to_class_descr(select):
	
	select, bracketed = remove_bracketed_all(select)
	select = select.split(".")
	if len(select) != 2:
		return None
	for i in range(2):
		for key in bracketed:
			select[i] = select[i].replace(key, bracketed[key][1:-1])
	return select

def load_user_tool(path):
	
	def _replace_bracketed(substr, bracketed):
		
		for key in bracketed:
			substr = substr.replace(key, bracketed[key][1:-1])
		return substr
	
	markup = ""
	with open(path, "r", encoding = "utf-8") as f:
		markup = f.read()
	
	markup, bracketed = remove_bracketed_all(markup)
	
	elements = []
	# elements = [["Title", title], ["Type", type], tag, group, multigroup, select, unique, ...]
	#	type = "Query" / "SearchForm" / "EntryForm"
	# 	tag = [control type, class.descriptor, label, stylesheet]
	# 	group = [["Group", label], tag, ...], multigroup = [["MultiGroup", label], tag, ...]
	#	select = [class.descriptor]
	collect = []
	idx0 = 0
	while True:
		idx = markup.find("<", idx0)
		value = markup[idx0:idx].strip()
		if value:
			collect[-1].append(value)
		idx0 = idx
		if idx0 == -1:
			break
		idx1 = markup.find(">", idx0)
		if idx1 == -1:
			break
		tag = markup[idx0+1:idx1].strip()
		if tag == "/":
			collect.append(-1)
		else:
			slash_end = tag.endswith("/")
			tag = tag.strip("/").strip()
			if tag:
				tag, quotes = remove_quoted(tag)
				stylesheet = ""
				if " style= _QU" in tag:
					idxs1 = tag.find(" style= _QU")
					idxs2 = tag.find("_ ", idxs1)
					if idxs2 == -1:
						idx0 = idx1 + 1
						continue
					key = tag[idxs1 + 7:idxs2 + 2].strip()
					stylesheet = quotes[key].strip("\"")
					del quotes[key]
					tag = tag[:idxs1]
				tag = [fragment.strip() % quotes for fragment in tag.split(" ")]
				tag = [fragment for fragment in tag if fragment]
				if tag:
					collect.append(tag + [stylesheet])
			if slash_end:
				collect.append(-1)
		idx0 = idx1 + 1
	
	while collect:
		tag = collect.pop(0)
		if not isinstance(tag, list):
			return
		if tag[0] in ["Group", "MultiGroup"]:
			group = [tag]
			while True:
				tag = collect.pop(0)
				if tag == -1:
					break
				group.append(tag.copy())
				if collect.pop(0) != -1:
					return
			elements.append(group)
		else:
			elements.append(tag.copy())
			if collect.pop(0) != -1:
				return
	
	if len(elements) < 3:
		return
	if elements[0][0] != "Title":
		return
	if elements[1][0] != "Type":
		return
	
	data = dict(
		typ = elements[1][1],
		label = elements[0][2],
		elements = []
	)
	if elements[1][1] == "Query":
		if elements[2][0] != "QueryString":
			return
		data["value"] = elements[2][2]
	else:
		for element in elements[2:]:
			if element[0] == "ColumnBreak":
				data["elements"].append(dict(
					typ = "ColumnBreak",
				))
			elif element[0] == "Select":
				dclass, descriptor = element[1].split(".")
				dclass = _replace_bracketed(dclass, bracketed)
				descriptor = _replace_bracketed(descriptor, bracketed)
				data["elements"].append(dict(
					typ = "Select",
					dclass = dclass,
					descriptor = descriptor,
					label = "",
					stylesheet = "",
				))
			elif element[0] == "Unique":
				dclass = element[1]
				data["elements"].append(dict(
					typ = "Unique",
					dclass = dclass,
					label = "",
					stylesheet = "",
				))
			elif isinstance(element[0], list): # Group or MultiGroup
				data["elements"].append(dict(
					typ = element[0][0],
					stylesheet = element[0][1],
					label = element[0][2],
					members = [],
				))
				for typ, select, stylesheet, label in element[1:]:
					dclass, descriptor = select.split(".")
					dclass = _replace_bracketed(dclass, bracketed)
					descriptor = _replace_bracketed(descriptor, bracketed)
					data["elements"][-1]["members"].append(dict(
						typ = typ,
						dclass = dclass,
						descriptor = descriptor,
						label = label,
						stylesheet = stylesheet,
					))
			elif len(element) == 4:  # tag
				typ, select, stylesheet, label = element
				dclass, descriptor = select.split(".")
				dclass = _replace_bracketed(dclass, bracketed)
				descriptor = _replace_bracketed(descriptor, bracketed)
				data["elements"].append(dict(
					typ = typ,
					dclass = dclass,
					descriptor = descriptor,
					label = label,
					stylesheet = stylesheet,
				))
	
	return data



from deposit.store.dresource import DResource

from urllib.request import pathname2url, url2pathname, urlopen
from urllib.parse import urlparse, urljoin
from unidecode import unidecode
import validators
import tempfile
import filecmp
import imghdr
import shutil
import sys
import os

IMAGE_EXTENSIONS = ["svg", "png", "jpg", "jpeg", "gif", "tif", "tiff", "bmp"]
MAX_FILES_PER_DIR = 5000

def as_url(value):
	
	value = str(value)
	if is_local_url(value):
		return value
	if validators.url(value):
		return value
	value = os.path.normpath(os.path.abspath(value))
	if value.startswith("\\\\"):
		return "file:" + pathname2url(value)
	return urljoin("file:", pathname2url(value))


def is_local_url(url):
	
	return (urlparse(url).scheme == "file")


def url_to_path(url):
	
	parsed = urlparse(url)
	return url2pathname(parsed.path)


def get_temp_path(subdir = "temp", appdir = "deposit"):
	# return temporary path
	
	tempdir = os.path.normpath(os.path.abspath(os.path.join(tempfile.gettempdir(), appdir, subdir)))
	if not os.path.isdir(tempdir):
		try:
			os.makedirs(tempdir)
		except:
			print("Dir not created: %s" % tempdir)
	return tempdir


def clear_temp_dir(appdir = "deposit"):
	
	tempdir = os.path.normpath(os.path.abspath(os.path.join(tempfile.gettempdir(), appdir)))
	if os.path.isdir(tempdir):
		shutil.rmtree(tempdir)


def get_named_path(name, local_folder):
	
	if local_folder is None:
		return get_temp_path(name)
	path = os.path.normpath(os.path.abspath(os.path.join(local_folder, name)))
	if not os.path.isdir(path):
		os.makedirs(path, exist_ok = True)
	return path


def sanitize_filename(filename, default = "file"):
	# return valid filename based on supplied filename, or default if not possible to sanitize
	
	def is_valid_filename(filename):
		# return True if filename is valid
		
		tmp_path = os.path.join(get_temp_path("is_valid_filename"), filename)
		try:
			f = open(tmp_path, "w")
		except:
			return False
		f.close()
		os.remove(tmp_path)
		return True
	
	if not filename:
		return default
	
	filename = unidecode(filename.strip())
	for i in range(len(filename)):
		if not (filename[i].isalnum() or (filename[i] in "._-")):
			filename = filename[:i] + "_" + filename[i+1:]
	if not is_valid_filename(filename):
		return default
	
	return filename


def extract_filename(url, default = "file"):
	# return filename extracted from path
	
	if is_local_url(url):
		url = url_to_path(url)
	
	filename = os.path.basename(url)
	if not filename:
		return default
	
	return filename


def get_unique_path(filename, folder, existing_urls = None):
	
	if existing_urls is None:
		existing_urls = set()
	existing_urls = set([os.path.basename(url) for url in existing_urls])
	
	name, ext = os.path.splitext(filename)
	filenames = set(os.listdir(folder))
	filenames.update(existing_urls)
	newname = name
	n = 1
	while newname + ext in filenames:
		newname = "%s%d" % (name, n)
		n += 1
	return os.path.normpath(os.path.abspath(os.path.join(folder, newname + ext)))


def get_image_format(url):
	# return format (extension) of local image file or None if not recognized
	
	if is_local_url(url):
		path = url_to_path(url)
		ext = os.path.splitext(path)[-1].lower().lstrip(".")
		if ext in IMAGE_EXTENSIONS:
			return ext
		return None
	format = None
	try:
		format = imghdr.what(None, urlopen(url).read())
	except:
		print("GET_IMAGE_FORMAT ERROR:", sys.exc_info())
	return format


def update_image_filename(filename, format):
	
	if not format:
		return filename
	name, ext = os.path.splitext(filename)
	ext = ext.strip()
	if ext == "":
		ext = "." + format
	return name + ext	


def get_free_subfolder(local_folder):
	
	found_dirs = {}
	dir_nums = set()
	for dirname in os.listdir(local_folder):
		if not dirname.isdigit():
			continue
		curdir = os.path.join(local_folder, dirname)
		if os.path.isdir(curdir):
			dir_nums.add(int(dirname))
			if (len(os.listdir(curdir)) < MAX_FILES_PER_DIR):
				found_dirs[int(dirname)] = curdir
	if found_dirs:
		return os.path.normpath(os.path.abspath(found_dirs[min(found_dirs.keys())]))
	
	n_max = max(dir_nums) + 1 if dir_nums else 0
	for n in range(n_max + 1):
		if n not in dir_nums:
			break
	folder = os.path.join(local_folder, "%04d" % (n))
	if not os.path.isdir(folder):
		os.makedirs(folder, exist_ok = True)
	return os.path.normpath(os.path.abspath(folder))


def get_updated_local_url(url, local_folder):
	
	if not is_local_url(url):
		return url
	
	local_folder = os.path.normpath(os.path.abspath(local_folder))
	
	path = url_to_path(url)
	if os.path.normpath(os.path.abspath(os.path.split(os.path.dirname(path))[0])) == local_folder:
		if os.path.isfile(path):
			return url
	
	filename = os.path.basename(path)
	for dirname in os.listdir(local_folder):
		if not dirname.isdigit():
			continue
		curdir = os.path.join(local_folder, dirname)
		path2 = os.path.join(curdir, filename)
		if os.path.isfile(path2):
			return as_url(path2)
	
	return None


def prune_local_files(except_paths, local_folder):
	
	del_path = get_named_path("_deleted", local_folder)
	for dirname in os.listdir(local_folder):
		if not dirname.isdigit():
			continue
		curdir = os.path.join(local_folder, dirname)
		for filename in os.listdir(curdir):
			path = os.path.join(curdir, filename)
			if True in [os.path.samefile(path, ex_path) for ex_path in except_paths]:
				continue
			tgt_path = os.path.join(del_path, filename)
			shutil.move(path, tgt_path)
			

def is_same_file(url1, url2):
	
	if not (is_local_url(url1) and is_local_url(url2)):
		return False
	
	return filecmp.cmp(url_to_path(url1), url_to_path(url2))


def store_locally(url, filename, local_folder, existing_urls = None):
	
	filename = sanitize_filename(filename)
	
	try:
		path_dst = get_unique_path(filename, get_free_subfolder(local_folder), existing_urls)
	except:
		print(sys.exc_info())
		return None
	
	if not copy_url(url, path_dst):
		return None
	
	return as_url(path_dst)


def delete_stored(url, local_folder):
	
	url = get_updated_local_url(url, local_folder)
	if url is None:
		return
	src_path = url_to_path(url)
	if not os.path.isfile(src_path):
		return False
	filename = os.path.basename(src_path)
	tgt_path0 = os.path.join(get_named_path("_deleted", local_folder), filename)
	path, filename = os.path.split(tgt_path0)
	name, ext = os.path.splitext(filename)
	tgt_path = tgt_path0
	n = 1
	while os.path.isfile(tgt_path):
		tgt_path = os.path.join(path, "%s_%d%s" % (name, n, ext))
		n += 1
	shutil.move(src_path, tgt_path)
	return True


def open_url(url):
	
	if is_local_url(url):
		path = url_to_path(url)
		return open(path, "rb")
	try:
		return urlopen(url).read()
	except:
		print("OPEN_URL ERROR:", sys.exc_info())
	return None


def copy_url(url_src, path_dst):
	
	if (not url_src) or (not path_dst):
		return False
	
	if is_local_url(url_src):
		try:
			shutil.copy2(url_to_path(url_src), path_dst)
			return True
		except:
			print("COPY_URL ERROR:", sys.exc_info())
		return False
	
	try:
		with urlopen(url_src) as response:
			with open(path_dst, "wb") as f:
				shutil.copyfileobj(response, f)
		return True
	except:
		print("COPY_URL ERROR:", sys.exc_info())
	
	return False


def copy_resources(resources, src_folder, dst_folder, progress = None):
	# resources = {url: DResource, ...}
	
	cmax = len(resources)
	cnt = 1
	collect = {}
	for url in resources:
		if progress is not None:
			if progress.cancel_pressed():
				return None
			progress.update_state(value = cnt, maximum = cmax)
		cnt += 1
		
		url, filename, is_stored, is_image = resources[url].value
		if is_stored:
			url = get_updated_local_url(url, src_folder)
			url2 = get_updated_local_url(url, dst_folder)
			if is_same_file(url, url2):
				url = url2
			else:
				url = store_locally(url, filename, dst_folder, set(collect.keys()))
		collect[url] = DResource((url, filename, is_stored, is_image))
	
	return collect





from deposit.datasource.abstract_dbsource import AbstractDBSource

from deposit.utils.fnc_files import (is_local_url, url_to_path)
from deposit.utils.fnc_serialize import (GRAPH_ATTRS)

from collections import defaultdict
import json

class DB(AbstractDBSource):
	
	def __init__(self):
		
		AbstractDBSource.__init__(self)	
		
		self._identifiers = {}  # {identifier: [tablename, ...], ...}
	
	def connect(self):
		
		self._identifiers.clear()
		
		return AbstractDBSource.connect(self)
	
	def disconnect(self):
		
		self._identifiers.clear()
		
		return AbstractDBSource.disconnect(self)
	
	def get_current_tables(self):
		
		identifiers = self.get_identifiers()
		ident = self.get_identifier()
		if ident not in identifiers:
			return []
		return identifiers[ident]
	
	def is_deposit_db(self, tables = None):
		
		if tables is None:
			tables = self.get_current_tables()
		for name in [
			"db_meta", "user_tools", "queries", 
			"object_nodes", "object_relations", 
			"class_nodes", "class_relations", 
			"member_nodes", "member_relations",
			]:
			if name not in tables:
				return False
		return True
	
	def is_legacy_db(self, tables = None):
		
		if tables is None:
			tables = self.get_current_tables()
		for name in ["classes", "objects", "local_folder"]:
			if name not in tables:
				return False
		return True
	
	def get_identifiers(self):
		
		if (not self._identifiers):
			collect = defaultdict(list)
			for table in self.get_tables():
				ident = table.split("#")
				if len(ident) == 2:
					ident, name = ident
					collect[ident].append(name)
			self._identifiers = {}
			for ident in collect:
				if self.is_deposit_db(collect[ident]) or self.is_legacy_db(collect[ident]):
					self._identifiers[ident] = collect[ident].copy()
		
		return self._identifiers
	
	def is_valid(self):
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		ret = (cursor is not None) and (self._identifier in self.get_identifiers())
		
		if not was_connected:
			self.disconnect()
		
		return ret
	
	def delete(self):
		
		identifier = self.get_identifier()
		cursor = self.get_cursor()
		if (identifier is None) or (cursor is None):
			return False
		for table in self.get_current_tables():
			table = "%s#%s" % (identifier, table)
			cursor.execute("DROP TABLE IF EXISTS \"%s\";" % (table))
		cursor.connection.commit()
		self.disconnect()
		self.set_identifier(None)
		
		return True
	
	def load_data(self, progress = None):
		
		if self.is_legacy_db():
			return self.load_legacy_data(progress)
		
		identifier = self.get_identifier()
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		src_ = GRAPH_ATTRS["source"]
		tgt_ = GRAPH_ATTRS["target"]
		lbl_ = GRAPH_ATTRS["key"]
		id_ = GRAPH_ATTRS["name"]
		relations_ = GRAPH_ATTRS["link"]
		
		data = dict(
			object_relation_graph = dict(
				directed = True,
				multigraph = True,
				graph = {},
				nodes = [],
			),
			class_relation_graph = dict(
				directed = True,
				multigraph = True,
				graph = {},
				nodes = [],
			),
			class_membership_graph = dict(
				directed = True,
				multigraph = False,
				graph = {},
				nodes = [],
			),
			local_folder = None,
			max_order = None,
			user_tools = [],
			queries = {},
			deposit_version = None,
		)
		data["object_relation_graph"][relations_] = []
		data["class_relation_graph"][relations_] = []
		data["class_membership_graph"][relations_] = []
		
		table = identifier + "#db_meta"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		if rows:
			data_table = {}
			for row in rows:
				data_table[row[0]] = json.loads(row[1])
			data["deposit_version"] = data_table.get("deposit_version", None)
			data["local_folder"] = data_table.get("local_folder", None)
			data["max_order"] = data_table.get("max_order", None)
		
		table = identifier + "#object_nodes"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["object_relation_graph"]["nodes"].append(
				{id_: row[0], "data": json.loads(row[1])}
			)
		
		table = identifier + "#object_relations"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["object_relation_graph"][relations_].append({
				src_: row[0],
				tgt_: row[1],
				lbl_: row[2],
				"weight": row[3],
			})
		
		table = identifier + "#class_nodes"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["class_relation_graph"]["nodes"].append(
				{id_: row[0], "data": json.loads(row[1])}
			)
		
		table = identifier + "#class_relations"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["class_relation_graph"][relations_].append({
				src_: row[0],
				tgt_: row[1],
				lbl_: row[2],
			})
		
		table = identifier + "#member_nodes"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["class_membership_graph"]["nodes"].append(
				{id_: json.loads(row[0])}
			)
		
		table = identifier + "#member_relations"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["class_membership_graph"][relations_].append({
				src_: json.loads(row[0]),
				tgt_: json.loads(row[1]),
			})
		
		table = identifier + "#user_tools"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["user_tools"].append(json.loads(row[0]))
		
		table = identifier + "#queries"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		for row in rows:
			data["queries"][row[0]] = json.loads(row[1])
		
		if not was_connected:
			self.disconnect()
		
		return data
	
	def load_legacy_data(self, progress = None):
		# load data from postgres db (old format)
				
		identifier = self.get_identifier()
		tables = self.get_current_tables()
		
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		data = dict(
			classes = {},
			objects = {},
			user_tools = [],
			queries = {},
			local_folder = None,
		)
		
		cursor.execute("SELECT * FROM \"%s\";" % (identifier + "#local_folder",))
		for row in cursor.fetchall():
			local_folder = json.loads(row[0])
			if is_local_url(local_folder):
				local_folder = url_to_path(local_folder)
			data["local_folder"] = local_folder
			break
		
		table = identifier + "#classes"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		if rows:
			for row in rows:
				data["classes"][row[0]] = json.loads(row[1])
		
		table = identifier + "#objects"
		cursor.execute("SELECT * FROM \"%s\";" % (table,))
		rows = cursor.fetchall()
		if rows:
			for row in rows:
				data["objects"][row[0]] = json.loads(row[1])
		
		table = identifier + "#user_tools"
		if table in tables:
			cursor.execute("SELECT * FROM \"%s\";" % (table,))
			rows = cursor.fetchall()
			if rows:
				for row in rows:
					data["user_tools"].append(json.loads(row[0]))
		
		table = identifier + "#queries"
		if table in tables:
			cursor.execute("SELECT * FROM \"%s\";" % (table,))
			rows = cursor.fetchall()
			if rows:
				for row in rows:
					data["queries"][row[0]] = json.loads(row[1])
		
		if was_connected:
			self.disconnect()
		
		return data
	
	def save_data(self, 
		db_meta,
		object_nodes, object_relations, 
		class_nodes, class_relations, 
		member_nodes, member_relations, 
		user_tools, queries, 
		progress = None
	):
		# store data in postgres db
		'''
		db_meta = dict(
			deposit_version,
			local_folder,
			max_order,
		)
		object_nodes = [{id: int, data: object_node_data}, ...]
			object_node_data = {
				(optional) descriptors: {descr_name: value or AbstractDType.to_dict(), ...}
				(optional) locations: {descr_name: DGeometry.to_dict(), ...}
			}
		object_relations = [{src: int, tgt: int, lbl: str, weight: float}, ...]
		class_nodes = [{id: str, data: class_node_data}, ...]
			class_node_data = {
				order: int,
				(optional) descriptors = [descr_name, ...]
			}
		class_relations = [{src: str, tgt: str, lbl: str}, ...]
		member_nodes = [{id: int/str}, ...]
		member_relations = [{src: int/str, tgt: int/str}, ...]
		user_tools = [dict(), ...]
		queries = {name: querystr, ...}
		'''
		
		def create_table(name, columns, tables, cursor):

			if name.split("#")[-1] in tables:
				cursor.execute("DROP TABLE \"%s\";" % (name))
			cursor.execute("CREATE TABLE %s.\"%s\" (%s);" % (self.get_schema(), name, columns))
		
		identifier = self.get_identifier()
		tables = self.get_current_tables()
		was_connected = self.is_connected()
		cursor = self.get_cursor()
		
		src_ = GRAPH_ATTRS["source"]
		tgt_ = GRAPH_ATTRS["target"]
		lbl_ = GRAPH_ATTRS["key"]
		id_ = GRAPH_ATTRS["name"]
		
		table = identifier + "#db_meta"
		columns = "variable TEXT, value TEXT"
		create_table(table, columns, tables, cursor)
		if db_meta:
			data = []
			for variable in db_meta:
				data.append({"variable": variable, "value": json.dumps(db_meta[variable])})
			cursor.execute("""
				DROP TYPE IF EXISTS node_;
				CREATE TYPE node_ as (%s);
				INSERT INTO \"%s\" SELECT variable, value FROM json_populate_recordset(null::node_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#object_nodes"
		columns = "id INTEGER, data TEXT"
		create_table(table, columns, tables, cursor)
		if object_nodes:
			data = []
			for node in object_nodes:
				data.append({"id": node[id_], "data": json.dumps(node["data"])})
			cursor.execute("""
				DROP TYPE IF EXISTS node_;
				CREATE TYPE node_ as (%s);
				INSERT INTO \"%s\" SELECT id, data FROM json_populate_recordset(null::node_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#object_relations"
		columns = "src INTEGER, tgt INTEGER, label TEXT, weight FLOAT"
		create_table(table, columns, tables, cursor)
		if object_relations:
			data = []
			for rel in object_relations:
				data.append({"src": rel[src_], "tgt": rel[tgt_], "label": rel[lbl_], "weight": rel["weight"] if "weight" in rel else None})
			cursor.execute("""
				DROP TYPE IF EXISTS rel_;
				CREATE TYPE rel_ as (%s);
				INSERT INTO \"%s\" SELECT src, tgt, label, weight FROM json_populate_recordset(null::rel_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#class_nodes"
		columns = "id TEXT, data TEXT"
		create_table(table, columns, tables, cursor)
		if class_nodes:
			data = []
			for node in class_nodes:
				data.append({"id": node[id_], "data": json.dumps(node["data"])})
			cursor.execute("""
				DROP TYPE IF EXISTS node_;
				CREATE TYPE node_ as (%s);
				INSERT INTO \"%s\" SELECT id, data FROM json_populate_recordset(null::node_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#class_relations"
		columns = "src TEXT, tgt TEXT, label TEXT"
		create_table(table, columns, tables, cursor)
		if class_relations:
			data = []
			for rel in class_relations:
				data.append({"src": rel[src_], "tgt": rel[tgt_], "label": rel[lbl_]})
			cursor.execute("""
				DROP TYPE IF EXISTS rel_;
				CREATE TYPE rel_ as (%s);
				INSERT INTO \"%s\" SELECT src, tgt, label FROM json_populate_recordset(null::rel_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#member_nodes"
		columns = "id TEXT"
		create_table(table, columns, tables, cursor)
		if member_nodes:
			data = []
			for node in member_nodes:
				data.append({"id": json.dumps(node[id_])})
			cursor.execute("""
				DROP TYPE IF EXISTS node_;
				CREATE TYPE node_ as (%s);
				INSERT INTO \"%s\" SELECT id FROM json_populate_recordset(null::node_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#member_relations"
		columns = "src TEXT, tgt TEXT"
		create_table(table, columns, tables, cursor)
		if member_relations:
			data = []
			for rel in member_relations:
				data.append({"src": json.dumps(rel[src_]), "tgt": json.dumps(rel[tgt_])})
			cursor.execute("""
				DROP TYPE IF EXISTS rel_;
				CREATE TYPE rel_ as (%s);
				INSERT INTO \"%s\" SELECT src, tgt FROM json_populate_recordset(null::rel_, %%s);
			""" % (columns, table), (json.dumps(data),))		
		
		table = identifier + "#user_tools"
		columns = "data TEXT"
		create_table(table, columns, tables, cursor)
		if user_tools:
			data = []
			for row in user_tools:
				data.append({"data": json.dumps(row)})
			cursor.execute("""
				DROP TYPE IF EXISTS user_tool_;
				CREATE TYPE user_tool_ as (%s);
				INSERT INTO \"%s\" SELECT data FROM json_populate_recordset(null::user_tool_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		table = identifier + "#queries"
		columns = "title TEXT, querystr TEXT"
		create_table(table, columns, tables, cursor)
		if queries:
			data = []
			for title in queries:
				data.append({"title": title, "data": json.dumps(queries[title])})
			cursor.execute("""
				DROP TYPE IF EXISTS node_;
				CREATE TYPE node_ as (%s);
				INSERT INTO \"%s\" SELECT title, querystr FROM json_populate_recordset(null::node_, %%s);
			""" % (columns, table), (json.dumps(data),))
		
		cursor.connection.commit()
		
		if not was_connected:
			self.disconnect()
		
		return True

